__author__ = "Simon Nilsson"

from copy import deepcopy
from tkinter import *

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.pose_plotter import PosePlotter
from simba.plotting.pose_plotter_mp import PosePlotterMultiProcess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton,
                                        SimBADropDown, SimBALabel)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_int)
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.read_write import find_core_cnt, get_desktop_path, str_2_bool

ENTIRE_VIDEOS = "ENTIRE VIDEO(S)"
AUTO = 'AUTO'
KEYPOINT_SIZES = list(range(1, 101))
KEYPOINT_SIZES.insert(0, AUTO)
VIDEO_LENGTHS = list(range(10, 210, 10))
VIDEO_LENGTHS.insert(0, ENTIRE_VIDEOS)
COLOR_OPTIONS = deepcopy(Options.PALETTE_OPTIONS_CATEGORICAL.value)
COLOR_OPTIONS.insert(0, AUTO)


class VisualizePoseInFolderPopUp(PopUpMixin):

    """
    .. image:: _static/img/VisualizePoseInFolderPopUp.webp
       :width: 500
       :align: center

    :example:
    >>> VisualizePoseInFolderPopUp()
    """
    def __init__(self):

        PopUpMixin.__init__(self, title="VISUALIZE POSE ESTIMATION DATA", size=(800, 800), icon='pose')
        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_TOOLS.value)
        self.keypoint_size_dropdown = SimBADropDown(parent=self.settings_frm, label='KEY-POINT SIZES', label_width=30, dropdown_width=20, value=AUTO, command=None, dropdown_options=KEYPOINT_SIZES)
        self.video_slice_dropdown = SimBADropDown(parent=self.settings_frm, label="VIDEO SECTION (SECONDS):", label_width=30, dropdown_width=20, value=ENTIRE_VIDEOS, command=None, dropdown_options=VIDEO_LENGTHS)
        self.cpu_cnt_dropdown = SimBADropDown(parent=self.settings_frm, label="CPU COUNT:", label_width=30, dropdown_width=20, value=find_core_cnt()[1], command=None, dropdown_options=list(range(1, find_core_cnt()[0])))
        self.gpu_dropdown = SimBADropDown(parent=self.settings_frm, label="USE GPU:", label_width=30, dropdown_width=20, value='FALSE', command=None, dropdown_options=['TRUE', 'FALSE'])
        self.include_bbox_dropdown = SimBADropDown(parent=self.settings_frm, label="INCLUDE BOUNDING BOX:", label_width=30, dropdown_width=20, value='FALSE', command=None, dropdown_options=['TRUE', 'FALSE'])
        self.save_dir = FolderSelect(self.settings_frm, "SAVE DIRECTORY: ", title="Select a data folder", lblwidth=30, initialdir=get_desktop_path())
        self.number_of_animals_dropdown = SimBADropDown(parent=self.settings_frm, dropdown_options=list(range(1, 17)), label="NUMBER OF ANIMALS:", label_width=30, dropdown_width=20, command=self.__create_table, value=1)
        self.__create_table(animal_cnt=1)
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.keypoint_size_dropdown.grid(row=0, column=0, sticky=NW)
        self.video_slice_dropdown.grid(row=1, column=0, sticky=NW)
        self.cpu_cnt_dropdown.grid(row=2, column=0, sticky=NW)
        self.gpu_dropdown.grid(row=3, column=0, sticky=NW)
        self.include_bbox_dropdown.grid(row=4, column=0, sticky=NW)
        self.save_dir.grid(row=5, column=0, sticky=NW)
        self.number_of_animals_dropdown.grid(row=6, column=0, sticky=NW)

        self.main_frm.mainloop()

    def __create_table(self, animal_cnt):
        if hasattr(self, 'color_keypoint_frm'):
            self.color_keypoint_frm.destroy()
        animal_cnt = int(animal_cnt)
        dropdown_color_titles = [f'ANIMAL {x + 1} COLOR PALETTE:' for x in range(animal_cnt)]
        self.animal_clr_dropdowns = {}
        self.color_keypoint_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="ANIMAL COLORS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_TOOLS.value)
        for cnt, title in enumerate(dropdown_color_titles):
            self.animal_clr_dropdowns[title] = SimBADropDown(parent=self.color_keypoint_frm, dropdown_options=COLOR_OPTIONS, label=title, label_width=30, dropdown_width=20, value=COLOR_OPTIONS[0])
            self.animal_clr_dropdowns[title].grid(row=cnt, column=0, sticky=NW)
        self.color_keypoint_frm.grid(row=1, column=0, sticky=NW)

        self.single_video_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VISUALIZE SINGLE DATA FILE", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_TOOLS.value)
        self.single_instruction_lbl = SimBALabel(parent=self.single_video_frm, txt="SELECT CSV OR PARQUET FILE INSIDE 'PROJECT_FOLDER/CSV' SUB-DIRECTORIES", font=Formats.FONT_REGULAR_ITALICS.value)
        self.single_file_select = FileSelect(self.single_video_frm, "SELECT DATA FILE (CSV/PARQUET): ",  title="Select a data file", file_types=[("CSV or PARQUET", Options.WORKFLOW_FILE_TYPE_STR_OPTIONS.value)], lblwidth=40)
        self.single_file_run_btn = SimbaButton(parent=self.single_video_frm, txt='RUN', img='rocket', cmd=self.run, cmd_kwargs={'directory': lambda: False})
        self.single_video_frm.grid(row=2, column=0, sticky=NW)
        self.single_instruction_lbl.grid(row=0, column=0, sticky=NW)
        self.single_file_select.grid(row=1, column=0, sticky=NW)
        self.single_file_run_btn.grid(row=2, column=0, sticky=NW)

        self.video_dir_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="VISUALIZE DATA DIRECTORY", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.VIDEO_TOOLS.value)
        self.video_dir_lbl = SimBALabel(parent=self.video_dir_frm, txt="SELECT SUB-DIRECTORY INSIDE 'PROJECT_FOLDER/CSV' DIRECTORY", font=Formats.FONT_REGULAR_ITALICS.value)
        self.video_dir_select = FolderSelect(self.video_dir_frm, "SELECT DATA DIRECTORY (CSV/PARQUET): ",  title="Select a data folder", lblwidth=40)
        self.dir_run_btn = SimbaButton(parent=self.video_dir_frm, txt='RUN', img='rocket', cmd=self.run, cmd_kwargs={'directory': lambda: True})
        self.video_dir_frm.grid(row=3, column=0, sticky=NW)
        self.video_dir_lbl.grid(row=0, column=0, sticky=NW)
        self.video_dir_select.grid(row=1, column=0, sticky=NW)
        self.dir_run_btn.grid(row=2, column=0, sticky=NW)

    def run(self, directory: bool):
        if directory:
            data_path = self.video_dir_select.folder_path
            check_if_dir_exists(in_dir=data_path, source=self.__class__.__name__)
        else:
            data_path = self.single_file_select.file_path
            check_file_exist_and_readable(file_path=data_path)
        save_dir = self.save_dir.folder_path
        check_if_dir_exists(in_dir=save_dir, source=self.__class__.__name__)
        circle_size = self.keypoint_size_dropdown.getChoices()
        if not check_int(name='circle_size', value=circle_size, min_value=1, raise_error=False)[0]:
            circle_size = None
        else:
            circle_size = int(circle_size)
        gpu = str_2_bool(self.gpu_dropdown.get_value())
        bbox = str_2_bool(self.include_bbox_dropdown.get_value())
        cpu_cnt = int(self.cpu_cnt_dropdown.get_value())
        sample_time = self.video_slice_dropdown.getChoices()
        if sample_time == ENTIRE_VIDEOS: sample_time = None
        else: sample_time = int(sample_time)
        palettes = {}
        for cnt, (k, v) in enumerate(self.animal_clr_dropdowns.items()):
            if v.getChoices() == AUTO: palettes[cnt] = Options.PALETTE_OPTIONS_CATEGORICAL.value[cnt]
            else: palettes[cnt] = v.getChoices()
        if cpu_cnt == 1:
            plotter = PosePlotter(data_path=data_path,
                                  out_dir=save_dir,
                                  palettes=palettes,
                                  circle_size=circle_size,
                                  sample_time=sample_time)
        else:
            plotter = PosePlotterMultiProcess(data_path=data_path,
                                               out_dir=save_dir,
                                               palettes=palettes,
                                               gpu=gpu,
                                               circle_size=circle_size,
                                               sample_time=sample_time,
                                               bbox=bbox,
                                               core_cnt=cpu_cnt)
        plotter.run()

#VisualizePoseInFolderPopUp()


__author__ = "Simon Nilsson"

import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.roi_aggregate_statistics_analyzer import \
    ROIAggregateStatisticsAnalyzer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, SimbaButton, SimbaCheckbox)
from simba.utils.checks import check_float
from simba.utils.enums import Formats, Keys, Links
from simba.utils.errors import InvalidInputError, NoROIDataError

OUTSIDE_ROI = 'OUTSIDE REGIONS OF INTEREST'

class ROIAggregateDataAnalyzerPopUp(PopUpMixin, ConfigReader):
    """
    :example:
    >>> analyzer = ROIAggregateDataAnalyzerPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(msg='No ROI data detected in SimBA project. Draw ROIs before doing ROI time analysis', source=self.__class__.__name__)
        if len(self.outlier_corrected_paths) == 0:
            raise NoROIDataError(msg=f'No data found in {self.outlier_corrected_dir} directory. Create data before analyzing ROI time data', source=self.__class__.__name__)

        PopUpMixin.__init__(self, title="ROI AGGREGATE STATISTICS ANALYSIS", icon='data_table')
        self.config_path = config_path
        self.animal_cnt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMAL(S)", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.ROI_DATA_ANALYSIS.value)
        self.animal_cnt_dropdown = DropDownMenu(self.animal_cnt_frm, "# OF ANIMALS", list(range(1, self.animal_cnt + 1)), labelwidth=20)
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = SimbaButton(parent=self.animal_cnt_frm, txt="CONFIRM", img='tick', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self._get_settings_frm)
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self._get_settings_frm()

        self.main_frm.mainloop()

    def _get_settings_frm(self):
        if hasattr(self, "body_part_frm"):
            self.data_options_frm.destroy()
            self.body_part_frm.destroy()
            self.probability_frm.destroy()
            self.format_frm.destroy()

        self.probability_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT PROBABILITY THRESHOLD", icon_name='pose')
        self.probability_entry = Entry_Box(parent=self.probability_frm, fileDescription='PROBABILITY THRESHOLD (0.0-1.0):', labelwidth="30", value='0.0')
        self.probability_frm.grid(row=self.frame_children(frame=self.main_frm), sticky=NW)
        self.probability_entry.grid(row=0, column=0, sticky=NW)

        self.body_parts_dropdowns = {}
        self.body_part_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT BODY-PART(S)", icon_name='pose')
        self.body_part_frm.grid(row=self.frame_children(frame=self.main_frm), sticky=NW)
        for bp_cnt in range(int(self.animal_cnt_dropdown.getChoices())):
            self.body_parts_dropdowns[bp_cnt] = DropDownMenu(self.body_part_frm, f"BODY-PART {str(bp_cnt + 1)}:", self.body_parts_lst, "20")
            self.body_parts_dropdowns[bp_cnt].grid(row=bp_cnt, column=0, sticky=NW)
            self.body_parts_dropdowns[bp_cnt].setChoices(self.body_parts_lst[bp_cnt])

        self.data_options_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="DATA OPTIONS", icon_name='abacus')
        self.total_time_cb, self.total_time_var = SimbaCheckbox(parent=self.data_options_frm, txt='TOTAL ROI TIME (S)', val=True, txt_img='timer')
        self.entry_counts_cb, self.entry_counts_var = SimbaCheckbox(parent=self.data_options_frm, txt='ROI ENTRIES (COUNTS)', val=True, txt_img='abacus_2')
        self.first_entry_time_cb, self.first_entry_time_var = SimbaCheckbox(parent=self.data_options_frm, txt='FIRST ROI ENTRY TIME (S)', val=False, txt_img='one_blue')
        self.last_entry_time_cb, self.last_entry_time_var = SimbaCheckbox(parent=self.data_options_frm, txt='LAST ROI ENTRY TIME (S)', val=False, txt_img='finish')
        self.mean_bout_time_cb, self.mean_bout_time_var = SimbaCheckbox(parent=self.data_options_frm, txt='MEAN ROI BOUT TIME (S)', val=False, txt_img='average')
        self.detailed_cb, self.detailed_var = SimbaCheckbox(parent=self.data_options_frm, txt='DETAILED ROI BOUT DATA (SEQUENCES)', val=False, txt_img='details')
        self.movement_cb, self.movement_var = SimbaCheckbox(parent=self.data_options_frm, txt='ROI MOVEMENT (VELOCITY & DISTANCES)', val=False, txt_img='pose')
        self.outside_cb, self.outside_var = SimbaCheckbox(parent=self.data_options_frm, txt='OUTSIDE ROI ZONES DATA', val=False, txt_img='outside_2', tooltip_txt=f'TREAT ALL NON-ROI REGIONS AS AN ROI REGION NAMED \n "{OUTSIDE_ROI}"')

        self.data_options_frm.grid(row=self.frame_children(frame=self.main_frm), column=0, sticky=NW)
        self.total_time_cb.grid(row=0, column=0, sticky=NW)
        self.entry_counts_cb.grid(row=1, column=0, sticky=NW)
        self.first_entry_time_cb.grid(row=2, column=0, sticky=NW)
        self.last_entry_time_cb.grid(row=3, column=0, sticky=NW)
        self.mean_bout_time_cb.grid(row=4, column=0, sticky=NW)
        self.detailed_cb.grid(row=5, column=0, sticky=NW)
        self.movement_cb.grid(row=6, column=0, sticky=NW)
        self.outside_cb.grid(row=7, column=0, sticky=NW)

        self.format_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="FORMAT OPTIONS", icon_name='abacus')
        self.transpose_cb, self.transpose_var = SimbaCheckbox(parent=self.format_frm, txt='TRANSPOSE OUTPUT TABLE', val=False, txt_img='restart')
        self.fps_cb, self.fps_var = SimbaCheckbox(parent=self.format_frm, txt='INCLUDE FPS DATA', val=False, txt_img='fps')
        self.video_length_cb, self.video_length_var = SimbaCheckbox(parent=self.format_frm, txt='INCLUDE VIDEO LENGTH DATA', val=False, txt_img='timer_2')
        self.px_per_mm_cb, self.px_per_mm_var = SimbaCheckbox(parent=self.format_frm, txt='INCLUDE PIXEL PER MILLIMETER DATA', val=False, txt_img='ruler')

        self.format_frm.grid(row=self.frame_children(frame=self.main_frm), column=0, sticky=NW)
        self.transpose_cb.grid(row=0, column=0, sticky=NW)
        self.fps_cb.grid(row=1, column=0, sticky=NW)
        self.video_length_cb.grid(row=2, column=0, sticky=NW)
        self.px_per_mm_cb.grid(row=3, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        #self.main_frm.mainloop()

    def run(self):
        probability = self.probability_entry.entry_get
        total_time = self.total_time_var.get()
        entry_counts = self.entry_counts_var.get()
        first_entry_time = self.first_entry_time_var.get()
        last_entry_time = self.last_entry_time_var.get()
        mean_bout_time = self.mean_bout_time_var.get()
        detailed_table = self.detailed_var.get()
        movement = self.movement_var.get()
        transpose = self.transpose_var.get()
        include_fps = self.fps_var.get()
        video_length = self.video_length_var.get()
        px_per_mm = self.px_per_mm_var.get()
        outside_roi = self.outside_var.get()

        body_parts = []
        for k, v in self.body_parts_dropdowns.items(): body_parts.append(v.getChoices())
        if not check_float(name='probability', value=probability, max_value=1.0, min_value=0.0, raise_error=False)[0]:
            raise InvalidInputError(msg=f'The PROBABILITY THRESHOLD has to be a value between 0.0 and 1.0 but got {probability}', source=self.__class__.__name__)
        selections = list({total_time, entry_counts, first_entry_time, last_entry_time, mean_bout_time, detailed_table, movement})
        if (len(selections) == 1) and (not selections[0]):
            raise InvalidInputError(msg=f'Please select at least one DATA OPTION.',source=self.__class__.__name__)

        analyzer =  ROIAggregateStatisticsAnalyzer(config_path=self.config_path,
                                                   data_path=None,
                                                   threshold=float(probability),
                                                   body_parts=body_parts,
                                                   detailed_bout_data=detailed_table,
                                                   calculate_distances=movement,
                                                   total_time=total_time,
                                                   entry_counts=entry_counts,
                                                   first_entry_time=first_entry_time,
                                                   last_entry_time=last_entry_time,
                                                   mean_bout_time=mean_bout_time,
                                                   transpose=transpose,
                                                   include_fps=include_fps,
                                                   include_video_length=video_length,
                                                   include_px_per_mm=px_per_mm,
                                                   outside_rois=outside_roi)

        analyzer.run()
        analyzer.save()


#_ = ROIAggregateDataAnalyzerPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")

#analyzer = ROIAggregateDataAnalyzerPopUp(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")





import os
from typing import Optional, Tuple, Union

import cv2
import numpy as np

from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_valid_img, check_if_valid_rgb_tuple,
                                check_int, check_str)
from simba.utils.enums import Options
from simba.utils.errors import InvalidFileTypeError
from simba.utils.read_write import get_fn_ext, read_frm_of_video
from simba.utils.warnings import CropWarning, ROIWarning


class ROISelectorCircle(object):
    """
    Class for selecting a circular region of interest (ROI) within an image or video frame.
    The selected region variables are stored in self: circle_center, circle_radius.

    :param Union[str, os.PathLike] path: Path to the image or video file. Can also be an image represented as a numpy array.
    :param int thickness: Thickness of the circle border when visualizing the ROI.
    :param Tuple[int, int, int] clr: BGR color tuple for visualizing the ROI. Default: deep pink.
    :param Optional[str] title: Title of the drawing window. If None, then `Draw ROI - Press ESC when drawn`.

    :raises InvalidFileTypeError: If the file type is not supported.
    :raises CropWarning: If the selected ROI extends beyond the image boundaries or if the ROI radius equals zero.

    .. image:: _static/img/circle_crop_2.gif
       :width: 700
       :align: center

    .. note::
       `Circle crop tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Tutorial_tools.md#circle-crop>`__.

    :example:
    >>> circle_selector = ROISelectorCircle(path='/Users/simon/Desktop/amber.png')
    >>> circle_selector.run()
    >>> circle_selector = ROISelectorCircle(path='/Users/simon/Desktop/Box2_IF19_7_20211109T173625_4_851_873_1_cropped.mp4')
    >>> circle_selector.run()
    >>> circle_selector = ROISelectorCircle(path=r"C:\troubleshooting\mitra\test\503_MA109_Gi_CNO_0521.mp4")
    >>> circle_selector.run()
    """

    def __init__(self,
                 path: Union[str, os.PathLike, np.ndarray],
                 thickness: int = 10,
                 clr: Tuple[int, int, int] = (147, 20, 255),
                 title: Optional[str] = None,
                 destroy: bool = True) -> None:

        check_if_valid_rgb_tuple(data=clr)
        check_int(name="Thickness", value=thickness, min_value=1, raise_error=True)
        if title is not None:
            check_str(name="ROI title window", value=title)
        if isinstance(path, np.ndarray):
            check_if_valid_img(path, source=self.__class__.__name__, raise_error=True)
            self.image = path
        else:
            check_file_exist_and_readable(file_path=path)
            _, filename, ext = get_fn_ext(filepath=path)
            if ext in Options.ALL_VIDEO_FORMAT_OPTIONS.value:
                self.image = read_frm_of_video(video_path=path)
            elif ext in Options.ALL_IMAGE_FORMAT_OPTIONS.value:
                self.image = cv2.imread(path)
            else:
                raise InvalidFileTypeError(
                    msg=f"Cannot crop a {ext} file.", source=self.__class__.__name__
                )
            if title is None:
                title = f"Draw ROI video {filename} - Press ESC when drawn"

        self.img_cpy = self.image.copy()
        self.w, self.h = self.image.shape[1], self.image.shape[0]
        if title is None:
            self.title = "Draw ROI - Press ESC when drawn"
        else:
            self.title = title

        self.destroy = destroy
        self.terminate = False
        self.img_copy = None
        self.drawing, self.clr, self.thickness = False, clr, thickness
        self.circle_center, self.circle_radius = (-1, -1), 1

    def draw_circle(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDOWN:
            self.drawing = True
            self.circle_center = (x, y)
        elif event == cv2.EVENT_MOUSEMOVE:
            if self.drawing:
                self.circle_radius = max(abs(x - self.circle_center[0]), abs(y - self.circle_center[1]))
        elif event == cv2.EVENT_LBUTTONUP:
            self.drawing = False
            self.circle_radius = max(abs(x - self.circle_center[0]), abs(y - self.circle_center[1]))
            self.img_copy = self.image.copy()
            cv2.circle(img=self.img_copy, center=self.circle_center, radius=self.circle_radius, color=self.clr, thickness=self.thickness)
            cv2.imshow(self.title, self.img_copy)

    def run(self):
        cv2.namedWindow(self.title, cv2.WINDOW_NORMAL)
        cv2.setMouseCallback(self.title, self.draw_circle)
        while not self.terminate:
            if self.drawing or self.img_copy is None:
                self.img_copy = self.image.copy()
                cv2.circle(self.img_copy, self.circle_center, self.circle_radius, self.clr, self.thickness )
            cv2.imshow(self.title, self.img_copy)
            key = cv2.waitKey(1)
            if key in [27, ord("q"), ord("Q"), ord(" ")]:
                if self.run_checks():
                    if self.destroy:
                        cv2.destroyAllWindows()
                    cv2.waitKey(1)
                    self.terminate = True
                    break

    def run_checks(self):
        self.left_border_tag = (int(self.circle_center[0] - self.circle_radius), self.circle_center[1])
        self.right_border_tag = (int(self.circle_center[0] + self.circle_radius), self.circle_center[1])
        self.top_border_tag = (self.circle_center[0], int(self.circle_center[1] - self.circle_radius))
        self.bottom_border_tag = (self.circle_center[0], int(self.circle_center[1] + self.circle_radius))
        if (((self.circle_center[0] - self.circle_radius) < 0) or ((self.circle_center[0] + self.circle_radius) > self.w) or ((self.circle_center[1] - self.circle_radius) < 0) or ((self.circle_center[1] + self.circle_radius) > self.h)):
            ROIWarning(msg="ROI WARNING: The drawn circular ROI radius extends beyond the image. Please try again.", source=self.__class__.__name__,)
            return False
        if self.circle_radius == 0:
            ROIWarning(msg="ROI WARNING: The drawn circular ROI radius equals 0. Please try again.", source=self.__class__.__name__)
            return False
        else:
            return True


# circle_selector = ROISelectorCircle(path=r"C:\troubleshooting\mitra\test\503_MA109_Gi_CNO_0521.mp4")
# circle_selector.run()


__author__ = "Simon Nilsson"

import os
import shutil
from datetime import datetime
from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_ffmpeg_available,
                                check_file_exist_and_readable, check_int,
                                check_nvidea_gpu_available, check_str,
                                check_valid_lst)
from simba.utils.enums import Paths, TagNames
from simba.utils.errors import FFMPEGCodecGPUError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import copy_files_to_directory, get_fn_ext
from simba.video_processors.video_processing import (
    horizontal_video_concatenator, mixed_mosaic_concatenator,
    mosaic_concatenator, vertical_video_concatenator)

HORIZONTAL = "horizontal"
VERTICAL = "vertical"
MOSAIC = "mosaic"
MIXED_MOSAIC = "mixed_mosaic"
ACCEPTED_TYPES = [HORIZONTAL, VERTICAL, MOSAIC, MIXED_MOSAIC]


class FrameMergererFFmpeg(ConfigReader):
    """
    Merge separate visualizations of classifications, descriptive statistics etc., into  single video mosaic.

    .. note::
       `GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-12-merge-frames>`_.

        .. image:: _static/img/mosaic_videos.gif
          :width: 600
          :align: center

    :parameter str config_path: Optional path to SimBA project config file in Configparser format.
    :parameter Literal["horizontal", "vertical", "mosaic", "mixed_mosaic"] concat_type: Type of concatenation. OPTIONS: 'horizontal', 'vertical', 'mosaic', 'mixed_mosaic'.
    :parameter List[Union[str, os.PathLike]] video_paths: List with videos to concatenate.
    :parameter Optional[int] video_height: Optional height of the canatenated videos. Required if concat concat_type is not mixed_mosaic.
    :parameter int video_width: Optional wisth of the canatenated videos. Required if concat concat_type is not mixed_mosaic.
    :parameter Optional[bool] gpu: If True, use NVIDEA FFMpeg GPU codecs. Default False.

    :example:
    >>> video_paths = ['/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT_downsampled.mp4', '/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT_downsampled.mp4']
    >>> merger = FrameMergererFFmpeg(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', video_paths=videos, video_height=600, video_width=600, concat_type='mosaic')
    >>> merger.run()
    """

    def __init__(
        self,
        concat_type: Literal["horizontal", "vertical", "mosaic", "mixed_mosaic"],
        video_paths: List[Union[str, os.PathLike]],
        video_height: Optional[int] = None,
        video_width: Optional[int] = None,
        config_path: Optional[str] = None,
        gpu: Optional[bool] = False,
    ):

        if gpu and not check_nvidea_gpu_available():
            raise FFMPEGCodecGPUError(
                msg="NVIDEA GPU not available (as evaluated by nvidea-smi returning None",
                source=self.__class__.__name__,
            )
        check_ffmpeg_available()
        check_str(
            name=f"{FrameMergererFFmpeg.__name__} concat_type",
            value=concat_type,
            options=ACCEPTED_TYPES,
        )
        check_valid_lst(
            data=video_paths,
            source=f"{self.__class__.__name__} video_paths",
            valid_dtypes=(str,),
            min_len=2,
        )
        for i in video_paths:
            check_file_exist_and_readable(file_path=i)
        if concat_type != MIXED_MOSAIC:
            check_int(
                name=f"{FrameMergererFFmpeg.__name__} video_height",
                value=video_height,
                min_value=0,
            )
            check_int(
                name=f"{FrameMergererFFmpeg.__name__} video_width",
                value=video_height,
                min_value=0,
            )
        if config_path is not None:
            ConfigReader.__init__(self, config_path=config_path)
            log_event(
                logger_name=str(__class__.__name__),
                log_type=TagNames.CLASS_INIT.value,
                msg=self.create_log_msg_from_init_args(locals=locals()),
            )
            self.output_dir = os.path.join(
                self.project_path, Paths.CONCAT_VIDEOS_DIR.value
            )
            self.output_path = os.path.join(
                self.project_path,
                Paths.CONCAT_VIDEOS_DIR.value,
                f"merged_video_{self.datetime}.mp4",
            )
        else:
            self.timer = SimbaTimer(start=True)
            self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")
            self.output_dir, _, _ = get_fn_ext(filepath=video_paths[0])
            self.output_path = os.path.join(
                self.output_dir, f"merged_video_{self.datetime}.mp4"
            )

        self.video_height, self.video_width, self.gpu = video_height, video_width, gpu
        self.video_paths, self.concat_type = video_paths, concat_type
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

    def run(self):
        if self.concat_type == HORIZONTAL:
            _ = horizontal_video_concatenator(
                video_paths=self.video_paths,
                save_path=self.output_path,
                height_px=self.video_height,
                gpu=self.gpu,
                verbose=True,
            )
        elif self.concat_type == VERTICAL:
            _ = vertical_video_concatenator(
                video_paths=self.video_paths,
                save_path=self.output_path,
                width_px=self.video_width,
                gpu=self.gpu,
                verbose=True,
            )
        elif self.concat_type == MOSAIC:
            _ = mosaic_concatenator(
                video_paths=self.video_paths,
                save_path=self.output_path,
                width_px=self.video_width,
                height_px=self.video_height,
                gpu=self.gpu,
                verbose=True,
            )
        else:
            _ = mixed_mosaic_concatenator(
                video_paths=self.video_paths,
                save_path=self.output_path,
                gpu=self.gpu,
                verbose=True,
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Merged video saved at {self.output_path}",
            source=self.__class__.__name__,
            elapsed_time=self.timer.elapsed_time_str,
        )


# videos = ['/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT_downsampled.mp4', '/Users/simon/Desktop/envs/simba/troubleshooting/mouse_open_field/project_folder/videos/SI_DAY3_308_CD1_PRESENT_downsampled.mp4']
#
# merger = FrameMergererFFmpeg(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                     video_paths=videos,
#                     video_height=600,
#                     video_width=600,
#                     concat_type='vertical') #horizontal, vertical, mosaic, mixed_mosaic
# merger.run()


#
# FrameMergererFFmpeg(config_path=None,
#                     frame_types={'Video 1': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                                  'Video 2': '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi'},
#                     video_height=640,
#                     video_width=480,
#                     concat_type='vertical') #horizontal, vertical, mosaic, mixed_mosaic
#
#


# FrameMergererFFmpeg(config_path=None,
#                     frame_types={'Video 1': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 2': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 3': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4',
#                                  'Video 4': r'C:\Users\Nape_Computer_2\Desktop\test_videos\Box1_PM2_day_5_20211104T171021.mp4'},
#                     video_height=640,
#                     video_width=480,
#                     concat_type='mixed_mosaic',
#                     gpu=False) #horizontal, vertical, mosaic, mixed_mosaic
#



__author__ = "Simon Nilsson"

import os.path
from tkinter import *
from typing import Union

from simba.feature_extractors.feature_subsets import FeatureSubsetsCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        FolderSelect, SimbaCheckbox)
from simba.utils.checks import check_if_dir_exists
from simba.utils.enums import Formats, Links
from simba.utils.errors import InvalidInputError, NoDataError

TWO_POINT_BP_DISTANCES = 'TWO-POINT BODY-PART DISTANCES (MM)'
WITHIN_ANIMAL_THREE_POINT_ANGLES = 'WITHIN-ANIMAL THREE-POINT BODY-PART ANGLES (DEGREES)'
WITHIN_ANIMAL_THREE_POINT_HULL = "WITHIN-ANIMAL THREE-POINT CONVEX HULL PERIMETERS (MM)"
WITHIN_ANIMAL_FOUR_POINT_HULL = "WITHIN-ANIMAL FOUR-POINT CONVEX HULL PERIMETERS (MM)"
ANIMAL_CONVEX_HULL_PERIMETER = 'ENTIRE ANIMAL CONVEX HULL PERIMETERS (MM)'
ANIMAL_CONVEX_HULL_AREA = "ENTIRE ANIMAL CONVEX HULL AREA (MM2)"
FRAME_BP_MOVEMENT = "FRAME-BY-FRAME BODY-PART MOVEMENTS (MM)"
FRAME_BP_TO_ROI_CENTER = "FRAME-BY-FRAME BODY-PART DISTANCES TO ROI CENTERS (MM)"
FRAME_BP_INSIDE_ROI = "FRAME-BY-FRAME BODY-PARTS INSIDE ROIS (BOOLEAN)"
ARENA_EDGE = "BODY-PART DISTANCES TO VIDEO FRAME EDGE (MM)"

FEATURE_FAMILIES = [TWO_POINT_BP_DISTANCES,
                    WITHIN_ANIMAL_THREE_POINT_ANGLES,
                    WITHIN_ANIMAL_THREE_POINT_HULL,
                    WITHIN_ANIMAL_FOUR_POINT_HULL,
                    ANIMAL_CONVEX_HULL_PERIMETER,
                    ANIMAL_CONVEX_HULL_AREA,
                    FRAME_BP_MOVEMENT,
                    FRAME_BP_TO_ROI_CENTER,
                    FRAME_BP_INSIDE_ROI,
                    ARENA_EDGE]

class FeatureSubsetExtractorPopUp(PopUpMixin, ConfigReader):
    """
    :example:
    >>> FeatureSubsetExtractorPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        if len(self.outlier_corrected_paths) == 0:
            raise NoDataError(msg=f'Cannot append feature subsets: No data found in {self.outlier_corrected_dir} directory.', source=self.__class__.__name__)
        PopUpMixin.__init__(self, title="EXTRACT FEATURE SUBSETS", size=(500, 500), icon='features')
        self.save_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="CHOOSE SAVE DIRECTORY", icon_name="save", icon_link=Links.FEATURE_SUBSETS.value)
        self.save_dir = FolderSelect(self.save_frm, "SAVE DIRECTORY:", lblwidth=20)
        self.save_frm.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=0, column=0, sticky=NW)

        self.append_settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="APPEND SETTINGS", icon_name="plus_green", icon_link=Links.FEATURE_SUBSETS.value)
        self.append_to_features_cb, self.append_to_features_var = SimbaCheckbox(parent=self.append_settings_frm, txt="APPEND RESULTS TO FEATURES EXTRACTED FILES")
        self.append_to_targets_cb, self.append_to_targets_var = SimbaCheckbox(parent=self.append_settings_frm, txt="APPEND RESULTS TO TARGET INSERTED FILES")
        self.checks_cb, self.checks_var = SimbaCheckbox(parent=self.append_settings_frm, txt="INCLUDE INTEGRITY CHECKS BEFORE APPENDING NEW DATA (RECOMMENDED)")
        self.append_settings_frm.grid(row=1, column=0, sticky=NW)
        self.append_to_features_cb.grid(row=0, column=0, sticky=NW)
        self.append_to_targets_cb.grid(row=1, column=0, sticky=NW)
        self.checks_cb.grid(row=2, column=0, sticky=NW)
        self.feature_subset_selections = self.create_cb_frame(main_frm=self.main_frm, cb_titles=FEATURE_FAMILIES, frm_title="SELECT FEATURE SUB-SETS",)
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        selected_features = []
        for feature_name, feature_var in self.feature_subset_selections.items():
            if feature_var.get(): selected_features.append(feature_name)
        if len(selected_features) == 0:
            raise InvalidInputError(msg="Please select at least ONE feature subset family.", source=self.__class__.__name__)
        if (not self.append_to_features_var.get() and not self.append_to_targets_var.get()):
            if not os.path.isdir(self.save_dir.folder_path):
                raise InvalidInputError(msg="You must select a valid path where to save the results OR select to append the features to the targets_inserted and/or features_extracted files.", source=self.__class__.__name__)
        if not os.path.isdir(self.save_dir.folder_path):
            save_dir = None
        else:
            save_dir = self.save_dir.folder_path
        feature_extractor = FeatureSubsetsCalculator(config_path=self.config_path,
                                                     feature_families=selected_features,
                                                     save_dir=save_dir,
                                                     file_checks=self.checks_var.get(),
                                                     append_to_features_extracted=self.append_to_features_var.get(),
                                                     append_to_targets_inserted=self.append_to_targets_var.get())
        feature_extractor.run()


#FeatureSubsetExtractorPopUp(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
from typing import Dict, Optional, Union

import numpy as np
import pandas as pd
from numba import jit

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import check_float, check_if_dir_exists, check_int
from simba.utils.enums import ConfigKey, Dtypes
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_config_entry, read_df,
                                    write_df)


def _movement_outlier_corrector(data_path: str,
                                config: ConfigReader,
                                animal_bp_dict: dict,
                                outlier_dict: dict,
                                file_type: str,
                                save_dir: str,
                                criterion: float):


    @jit(nopython=True)
    def _corrector(data: np.ndarray, criterion: float):
        results, current_value, cnt = np.full(data.shape, np.nan), data[0, :], 0
        for i in range(data.shape[0]):
            dist = abs(np.linalg.norm(current_value - data[i, :]))
            if dist <= criterion:
                current_value = data[i, :]
            else:
                cnt += 1
            results[i, :] = current_value
        return results, cnt


    def _outlier_replacer(data_df: pd.DataFrame,
                          animal_criteria: dict,
                          video_name: str):

        log = pd.DataFrame(columns=["VIDEO", "ANIMAL", "BODY-PART", "CORRECTION COUNT", "CORRECTION PCT"])
        for animal_name, animal_body_parts in animal_bp_dict.items():
            for bp_x_name, bp_y_name in zip(animal_body_parts["X_bps"], animal_body_parts["Y_bps"]):
                vals, cnt = _corrector(data=data_df[[bp_x_name, bp_y_name]].values, criterion=animal_criteria[animal_name])
                df = pd.DataFrame(vals, columns=[bp_x_name, bp_y_name])
                data_df.update(df)
                log.loc[len(log)] = [video_name, animal_name, bp_x_name[:-2], cnt, round(cnt / len(df), 6)]

        return data_df, log


    video_timer = SimbaTimer(start=True)
    _, video_name, _  = get_fn_ext(filepath=data_path)
    save_path = os.path.join(save_dir, f"{video_name}.{file_type}")
    df = read_df(data_path, file_type, check_multiindex=True)
    df = config.insert_column_headers_for_outlier_correction(data_df=df, new_headers=config.bp_headers, filepath=data_path)
    animal_criteria = {}
    for animal_name, animal_bps in outlier_dict.items():
        animal_bp_distances = np.sqrt((df[animal_bps["bp_1"] + "_x"] - df[animal_bps["bp_2"] + "_x"]) ** 2 + (df[animal_bps["bp_1"] + "_y"] - df[animal_bps["bp_2"] + "_y"]) ** 2)
        animal_criteria[animal_name] = (animal_bp_distances.mean() * criterion)
    df, log = _outlier_replacer(animal_criteria=animal_criteria, data_df=df, video_name=video_name)
    write_df(df=df, file_type=file_type, save_path=save_path)
    video_timer.stop_timer()
    print(f"Corrected movement outliers for file {video_name} (elapsed time: {video_timer.elapsed_time_str}s)...")

    return video_name, log



class OutlierCorrecterMovementMultiProcess(ConfigReader, FeatureExtractionMixin):
    """
    Detect and ammend outliers in pose-estimation data based on movement lenghth (Euclidean) of the body-parts
    in the current frame from preceeding frame. If not passed, then uses critera stored in the SimBA project project_config.ini
    under the [Outlier settings] headed. Uses multiprocessing.

    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Optional[Union[str, os.PathLike]] data_dir: The directory storing the input data. If None, then the ``input_csv`` directory of the SimBA project.
    :param Optional[Union[str, os.PathLike]] save_dir: The directory to store the results. If None, then the ``outlier_corrected_movement`` directory of the SimBA project.
    :param Optional[int] core_cnt: The number of cores to use. If -1, then all available cores. Default: -1.
    :param Optional[Dict[str, Dict[str, str]]] animal_dict: Dictionary holding the animal names, and the two body-parts to use to measure the mean or median size of the animals. If None, grabs the info from the SimBA project config.
    :param Optional[float] criterion: The criterion multiplier. If None, grabs the info from the SimBA project config.

    .. image:: _static/img/movement_outlier.png
       :width: 500
       :align: center

    .. note::
       `Outlier correction documentation <https://github.com/sgoldenlab/simba/blob/master/misc/Outlier_settings.pdf>`__.

    :example:
    >>> outlier_correcter_movement = OutlierCorrecterMovementMultiProcess(config_path='MyProjectConfig')
    >>> outlier_correcter_movement.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_dir: Optional[Union[str, os.PathLike]] = None,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 core_cnt: Optional[int] = -1,
                 animal_dict: Optional[Dict[str, Dict[str, str]]] = None,
                 criterion: Optional[float] = None):

        ConfigReader.__init__(self, config_path=config_path, create_logger=False)
        FeatureExtractionMixin.__init__(self)
        if not os.path.exists(self.outlier_corrected_movement_dir):
            os.makedirs(self.outlier_corrected_movement_dir)
        if criterion is None:
            self.criterion = read_config_entry(self.config, ConfigKey.OUTLIER_SETTINGS.value, ConfigKey.MOVEMENT_CRITERION.value, Dtypes.FLOAT.value)
        else:
            check_float(name=f'{criterion} criterion', value=criterion, min_value=10e-10)
            self.criterion = criterion
        if data_dir is not None:
            check_if_dir_exists(in_dir=data_dir, source=self.__class__.__name__)
            self.data_dir = data_dir
        else:
            self.data_dir = self.input_csv_dir

        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=self.__class__.__name__)
            self.save_dir = save_dir
        else:
            self.save_dir = self.outlier_corrected_movement_dir

        check_int(name=f'{self.__class__.__name__} core_cnt', value=core_cnt, min_value=-1, unaccepted_vals=[0])
        self.core_cnt = core_cnt
        if self.core_cnt == -1:
            self.core_cnt = find_core_cnt()[0]

        self.outlier_bp_dict, self.above_criterion_dict_dict = {}, {}
        if animal_dict is None:
            if self.animal_cnt == 1:
                self.animal_id = read_config_entry(self.config, ConfigKey.MULTI_ANIMAL_ID_SETTING.value, ConfigKey.MULTI_ANIMAL_IDS.value, Dtypes.STR.value)
                if self.animal_id != "None":
                    self.animal_bp_dict[self.animal_id] = self.animal_bp_dict.pop("Animal_1")

            for animal_name in self.animal_bp_dict.keys():
                self.outlier_bp_dict[animal_name] = {}
                self.outlier_bp_dict[animal_name]["bp_1"] = read_config_entry(self.config,"Outlier settings", "movement_bodypart1_{}".format(animal_name.lower()),"str")
                self.outlier_bp_dict[animal_name]["bp_2"] = read_config_entry(self.config,"Outlier settings", "movement_bodypart2_{}".format(animal_name.lower()),"str")
        else:
            self.outlier_bp_dict = animal_dict

    def run(self):
        self.logs = []
        data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=True)
        data_path_tuples = [(x) for x in data_paths]
        with multiprocessing.Pool(self.core_cnt, maxtasksperchild=self.maxtasksperchild) as pool:
            constants = functools.partial(_movement_outlier_corrector,
                                          config=self,
                                          animal_bp_dict=self.animal_bp_dict,
                                          outlier_dict=self.outlier_bp_dict,
                                          save_dir=self.save_dir,
                                          file_type=self.file_type,
                                          criterion=self.criterion)
            for cnt, (video_name, log) in enumerate(pool.imap(constants, data_path_tuples, chunksize=1)):
                print(f"Video {video_name} complete...")
                self.logs.append(log)

        self.__save_log_file()

    def __save_log_file(self):
        log_fn = os.path.join(self.logs_path, f"Outliers_movement_{self.datetime}.csv")
        self.logs = pd.concat(self.logs, axis=0)
        self.logs.to_csv(log_fn)
        self.timer.stop_timer()
        stdout_success(msg=f'Log for corrected "movement outliers" saved in {self.logs_path}', elapsed_time=self.timer.elapsed_time_str)

#
# if __name__ == "__main__":
#     #test = OutlierCorrecterMovementMultiProcess(config_path=r"C:\troubleshooting\two_black_animals_14bp\project_folder\project_config.ini")
#     test = OutlierCorrecterMovementMultiProcess(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini")
#     test.run()
#



import math
from tkinter import *
from typing import List, Optional, Tuple

import cv2
import numpy as np
import pandas as pd
from PIL import Image, ImageTk
from shapely.geometry import Polygon

from simba.mixins.plotting_mixin import PlottingMixin
from simba.roi_tools.roi_utils import (get_circle_df_headers,
                                       get_image_from_label,
                                       get_polygon_df_headers,
                                       get_rectangle_df_headers,
                                       insert_gridlines_on_roi_img)
from simba.utils.checks import check_instance, check_valid_polygon
from simba.utils.enums import ROI_SETTINGS, Keys

DRAW_FRAME_NAME = "DEFINE SHAPE"

BR_TAG = 'Bottom right tag'
B_TAG = 'Bottom tag'
T_TAG = 'Top tag'
C_TAG = 'Center tag'
BL_TAG = 'Bottom left tag'
TR_TAG = 'Top right tag'
TL_TAG = 'Top left tag'
R_TAG = 'Right tag'
L_TAG = 'Left tag'
BR_X = "Bottom_right_X"
BR_Y = "Bottom_right_Y"
TL_X = 'topLeftX'
TL_Y = "topLeftY"
CENTER_X, CENTER_Y = "Center_X", "Center_Y"
HEIGHT, WIDTH = 'height', 'width'
BORDER_TAG = 'Border tag'
CIRCLE_C_X = 'centerX'
CIRCLE_C_Y = 'centerY'
RADIUS = 'radius'
VERTICES = 'vertices'
OVERLAY_GRID_COLOR = 'OVERLAY_GRID_COLOR'


def _plot_roi(roi_dict: dict,
              img: np.ndarray,
              show_tags: bool = False,
              omitted_roi: Optional[str] = None):
    rectangles_df, circles_df, polygon_df = pd.DataFrame(columns=get_rectangle_df_headers()), pd.DataFrame(
        columns=get_circle_df_headers()), pd.DataFrame(columns=get_polygon_df_headers())
    for roi_name, roi_data in roi_dict.items():
        if (roi_data['Shape_type'].lower() == ROI_SETTINGS.RECTANGLE.value) and (roi_name != omitted_roi):
            rectangles_df = pd.concat([rectangles_df, pd.DataFrame([roi_data])], ignore_index=True)
        elif roi_data['Shape_type'].lower() == ROI_SETTINGS.CIRCLE.value and (roi_name != omitted_roi):
            circles_df = pd.concat([circles_df, pd.DataFrame([roi_data])], ignore_index=True)
        elif roi_data['Shape_type'].lower() == ROI_SETTINGS.POLYGON.value and (roi_name != omitted_roi):
            polygon_df = pd.concat([polygon_df, pd.DataFrame([roi_data])], ignore_index=True)
    roi_dict = {Keys.ROI_RECTANGLES.value: rectangles_df, Keys.ROI_CIRCLES.value: circles_df,
                Keys.ROI_POLYGONS.value: polygon_df}
    img = PlottingMixin.roi_dict_onto_img(img=img, roi_dict=roi_dict, circle_size=None, show_tags=show_tags)
    return img


class InteractiveROIModifier():

    def __init__(self,
                 img_window: Toplevel,
                 original_img: np.ndarray,
                 roi_dict: dict,
                 settings: Optional[dict] = None,
                 hex_grid: Optional[List[Polygon]] = None,
                 rectangle_grid: Optional[List[Polygon]] = None):

        check_instance(source=self.__class__.__name__, instance=img_window, accepted_types=(Toplevel,))
        if settings is None: settings = {item.name: item.value for item in ROI_SETTINGS}
        self.hex_grid, self.rectangle_grid = hex_grid, rectangle_grid
        self.img_lbl = img_window.nametowidget("img_lbl")
        self.img = get_image_from_label(self.img_lbl)
        self.original_img, self.roi_dict = original_img, roi_dict
        self.img_w, self.img_h = self.img.shape[1], self.img.shape[0]
        self.circle_size = PlottingMixin().get_optimal_circle_size(frame_size=(self.img_w, self.img_h), circle_frame_ratio=100)
        self.img_window, self.edge_selected, self.settings = img_window, False, settings
        self.bind_keys()

    def bind_keys(self):
        self.img_window.bind("<ButtonPress-1>", self.left_mouse_down)
        self.img_window.bind("<B1-Motion>", self.left_mouse_drag)
        self.img_window.bind("<ButtonRelease-1>", self.left_mouse_up)

    def unbind_keys(self):
        self.img_window.unbind("<ButtonPress-1>")
        self.img_window.unbind("<B1-Motion>")
        self.img_window.unbind("<ButtonRelease-1>")

    def find_closest_tag(self, roi_dict: dict, click_coordinate: Tuple[int, int]):
        clicked_roi, clicked_tag = None, None
        for roi_name, roi_data in roi_dict.items():
            ear_tag_size = roi_data['Ear_tag_size']
            for roi_tag_name, roi_tag_coordinate in roi_data['Tags'].items():
                distance = math.sqrt((roi_tag_coordinate[0] - click_coordinate[0]) ** 2 + (roi_tag_coordinate[1] - click_coordinate[1]) ** 2)
                if distance <= ear_tag_size:
                    clicked_roi, clicked_tag = roi_data, roi_tag_name
        return clicked_roi, clicked_tag


    def get_gridline_copy(self):
        gridline_img = self.original_img.copy()
        if self.rectangle_grid is not None:
            gridline_img = insert_gridlines_on_roi_img(img=gridline_img, grid=self.rectangle_grid, color=self.settings[OVERLAY_GRID_COLOR], thickness=max(1, int(self.circle_size/5)))
        if self.hex_grid is not None:
            gridline_img = insert_gridlines_on_roi_img(img=gridline_img, grid=self.rectangle_grid, color=self.settings[OVERLAY_GRID_COLOR], thickness=max(1, int(self.circle_size / 5)))
        return gridline_img

    def update_image(self, img):
        img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        pil_image = Image.fromarray(img_rgb)
        tk_image = ImageTk.PhotoImage(pil_image)
        self.img_lbl.configure(image=tk_image)
        self.img_lbl.image = tk_image


    def _select_rectangle_bottom_right(self):
        self.clicked_roi[BR_X], self.clicked_roi[BR_Y] = self.x, self.y
        self.clicked_roi[HEIGHT], self.clicked_roi[WIDTH] = self.h, self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = int(self.tl_x + self.w / 2), int(self.tl_y + self.h / 2)
        self.clicked_roi['Tags'][TR_TAG] = (self.x, self.y - self.h)
        self.clicked_roi['Tags'][BL_TAG] = (self.x - self.w, self.y)
        self.clicked_roi['Tags'][BR_TAG] = (self.x, self.y)
        self.clicked_roi['Tags'][R_TAG] = (self.x, int(self.tl_y + (self.h / 2)))
        self.clicked_roi['Tags'][L_TAG] = (self.tl_x, int(self.tl_y + (self.h / 2)))
        self.clicked_roi['Tags'][B_TAG] = (int(self.tl_x + self.w / 2), self.y)
        self.clicked_roi['Tags'][T_TAG] = (int(self.tl_x + self.w / 2), self.y - self.h)
        self.clicked_roi['Tags'][C_TAG] = (self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y])
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], (self.x, self.y), self.clicked_roi['Color BGR'],self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, (self.x, self.y), self.clicked_roi['Tags'][BL_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.temp_img = cv2.line(self.temp_img, (self.x, self.y), self.clicked_roi['Tags'][TR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_right(self):
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = int(self.tl_x + self.w / 2), int(self.tl_y + self.h / 2)
        self.clicked_roi['Tags'][TR_TAG] = (self.x, self.tl_y)
        self.clicked_roi['Tags'][BR_TAG] = (self.x, self.tl_y + self.h)
        self.clicked_roi['Tags'][R_TAG] = (self.tl_x + self.w, self.tl_y + int(self.h / 2))
        self.clicked_roi['Tags'][T_TAG] = (self.tl_x + int(self.w / 2), self.tl_y)
        self.clicked_roi['Tags'][B_TAG] = (self.tl_x + int(self.w / 2), self.tl_y + self.h)
        self.clicked_roi['Tags'][C_TAG] = (self.tl_x + int(self.w / 2), self.tl_y + int(self.h / 2))
        self.clicked_roi[BR_X] = self.x
        self.clicked_roi[BR_Y] = self.tl_y + self.h
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img =  cv2.line(self.temp_img, self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Tags'][TR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_bottom(self):
        self.clicked_roi['Tags'][BR_TAG] = (self.tl_x + self.w, self.tl_y + self.h)
        self.clicked_roi['Tags'][BL_TAG] = (self.tl_x, self.tl_y + self.h)
        self.clicked_roi['Tags'][L_TAG] = (self.tl_x, self.tl_y + int(self.h / 2))
        self.clicked_roi['Tags'][R_TAG] = (self.tl_x + self.w, self.tl_y + int(self.h / 2))
        self.clicked_roi['Tags'][B_TAG] = (self.tl_x + int(self.w / 2), self.tl_y + self.h)
        self.clicked_roi['Tags'][C_TAG] = (self.tl_x + int(self.w / 2), self.tl_y + int(self.h / 2))
        self.clicked_roi[BR_X] = self.tl_x + self.w
        self.clicked_roi[BR_Y] = self.tl_y + self.h
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.tl_x + int(self.w / 2), self.tl_y + int(self.h / 2)
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][BL_TAG], self.clicked_roi['Tags'][BR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_bottom_left(self):
        self.clicked_roi['Tags'][TL_TAG] = (self.x, self.tr_y)
        self.clicked_roi['Tags'][L_TAG] = (self.x, self.tr_y + int(self.h / 2))
        self.clicked_roi['Tags'][BL_TAG] = (self.x, self.y)
        self.clicked_roi['Tags'][B_TAG] = (self.x + int(self.w / 2), self.y)
        self.clicked_roi['Tags'][BR_TAG] = (self.x + self.w, self.tr_y + self.h)
        self.clicked_roi['Tags'][R_TAG] = (self.x + self.w, self.tr_y + int(self.h / 2))
        self.clicked_roi['Tags'][T_TAG] = (self.x + int(self.w / 2), self.tr_y)
        self.clicked_roi['Tags'][C_TAG] = (self.x + int(self.w / 2), self.tr_y + int(self.h / 2))
        self.clicked_roi[TL_X] = self.x
        self.clicked_roi[BR_X] = self.x + self.w
        self.clicked_roi[BR_Y] = self.tr_y + self.h
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.x + int(self.w / 2), self.tr_y + int(self.h / 2)
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][BL_TAG], self.clicked_roi['Tags'][TL_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][BL_TAG], self.clicked_roi['Tags'][BR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_left_tag(self):
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = int(self.tr_x - (self.w / 2)), int(self.tr_y + self.h / 2)
        self.clicked_roi['Tags'][BL_TAG] = (self.tr_x - self.w, self.tr_y + self.h)
        self.clicked_roi['Tags'][TL_TAG] = (self.tr_x - self.w, self.tr_y)
        self.clicked_roi['Tags'][L_TAG] = (self.x, int(self.tr_y + self.h / 2))
        self.clicked_roi['Tags'][B_TAG] = (self.tr_x - int(self.w / 2), self.tr_y + self.h)
        self.clicked_roi['Tags'][T_TAG] = (self.tr_x - int(self.w / 2), self.tr_y)
        self.clicked_roi['Tags'][C_TAG] = (int(self.tr_x - (self.w / 2)), int(self.tr_y + self.h / 2))
        self.clicked_roi[TL_X] = self.tr_x - self.w
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][BL_TAG], self.clicked_roi['Tags'][TL_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_top_tag(self):
        self.clicked_roi['Tags'][TR_TAG] = (self.br_x, self.br_y - self.h)
        self.clicked_roi['Tags'][TL_TAG] = (self.br_x - self.w, self.br_y - self.h)
        self.clicked_roi['Tags'][T_TAG] = (self.br_x - int(self.w / 2), self.br_y - self.h)
        self.clicked_roi['Tags'][L_TAG] = (self.br_x - self.w, self.br_y - int(self.h / 2))
        self.clicked_roi['Tags'][R_TAG] = (self.br_x, self.br_y - int(self.h / 2))
        self.clicked_roi['Tags'][C_TAG] = (self.br_x - int(self.w / 2), self.br_y - int(self.h / 2))
        self.clicked_roi[TL_X] = self.br_x - self.w
        self.clicked_roi[TL_Y] = self.br_y - self.h
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.br_x - int(self.w / 2), self.br_y - int(self.h / 2)
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][TR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_top_left_tag(self):
        self.clicked_roi['Tags'][TR_TAG] = (self.br_x, self.br_y - self.h)
        self.clicked_roi['Tags'][T_TAG] = (self.br_x - int(self.w / 2), self.br_y - self.h)
        self.clicked_roi['Tags'][L_TAG] = (self.br_x - self.w, self.br_y - int(self.h / 2))
        self.clicked_roi['Tags'][R_TAG] = (self.br_x, self.br_y - int(self.h / 2))
        self.clicked_roi['Tags'][BL_TAG] = (self.x, self.br_y)
        self.clicked_roi['Tags'][B_TAG] = (self.x + int(self.w/2), self.br_y)
        self.clicked_roi['Tags'][TL_TAG] = (self.x, self.y)
        self.clicked_roi['Tags'][C_TAG] = (self.x + int(self.w / 2), self.y + int(self.h / 2))
        self.clicked_roi[TL_X] = self.x
        self.clicked_roi[TL_Y] = self.y
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.x + int(self.w / 2), self.y + int(self.h / 2)
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][TR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BL_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_top_right_tag(self):
        self.clicked_roi['Tags'][TL_TAG] = (self.bl_x, self.y)
        self.clicked_roi['Tags'][L_TAG] = (self.bl_x, self.bl_y - int(self.h / 2))
        self.clicked_roi['Tags'][T_TAG] = (self.bl_x + int(self.w / 2), self.y)
        self.clicked_roi['Tags'][TR_TAG] = (self.x, self.y)
        self.clicked_roi['Tags'][R_TAG] = (self.x, self.bl_y - int(self.h / 2))
        self.clicked_roi['Tags'][BR_TAG] = (self.x, self.y + self.h)
        self.clicked_roi['Tags'][B_TAG] = (self.x - int(self.w / 2), self.bl_y)
        self.clicked_roi['Tags'][C_TAG] = (self.bl_x + int(self.w / 2), self.bl_y - int(self.h / 2))
        self.clicked_roi[TL_X] = self.bl_x
        self.clicked_roi[TL_Y] = self.y
        self.clicked_roi[BR_X] = self.x
        self.clicked_roi[BR_Y] = self.y + self.h
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.bl_x + int(self.w / 2), self.bl_y - int(self.h / 2)
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.clicked_roi['Color BGR'], self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][TR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][TR_TAG], self.clicked_roi['Tags'][BR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_rectangle_center_tag(self):
        self.clicked_roi['Tags'][TL_TAG] = (self.tl_x, self.tl_y)
        self.clicked_roi['Tags'][L_TAG] = (self.tl_x, self.tl_y + int(self.h / 2))
        self.clicked_roi['Tags'][BL_TAG] = (self.tl_x, self.tl_y + self.h)
        self.clicked_roi['Tags'][B_TAG] = (self.tl_x + int(self.w / 2), self.tl_y + self.h)
        self.clicked_roi['Tags'][BR_TAG] = (self.br_x, self.br_y)
        self.clicked_roi['Tags'][R_TAG] = (self.tl_x + self.w, self.tl_y + int(self.h / 2))
        self.clicked_roi['Tags'][TR_TAG] = (self.tl_x + self.w, self.tl_y)
        self.clicked_roi['Tags'][T_TAG] = (self.tl_x + int(self.w / 2), self.tl_y)
        self.clicked_roi['Tags'][C_TAG] = (self.x, self.y)
        self.clicked_roi[TL_X] = self.tl_x
        self.clicked_roi[TL_Y] = self.tl_y
        self.clicked_roi[BR_X] = self.br_x
        self.clicked_roi[BR_Y] = self.br_y
        self.clicked_roi[HEIGHT] = self.h
        self.clicked_roi[WIDTH] = self.w
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.x, self.y
        self.temp_img = cv2.rectangle(self.temp_img, self.clicked_roi['Tags'][TL_TAG], self.clicked_roi['Tags'][BR_TAG], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_circle_border(self):
        self.clicked_roi[RADIUS] = self.radius
        self.clicked_roi["Tags"][BORDER_TAG] = self.l_edge
        self.temp_img = cv2.circle(self.temp_img, center=self.clicked_roi["Tags"][C_TAG], radius=self.radius, color=self.settings['ROI_SELECT_CLR'], thickness=self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_circle_center(self):
        self.clicked_roi[CIRCLE_C_X], self.clicked_roi[CIRCLE_C_Y] = self.x, self.y
        self.clicked_roi["Tags"][C_TAG] = (self.x, self.y)
        self.clicked_roi["Tags"][BORDER_TAG] = self.l_edge
        self.temp_img = cv2.circle(self.temp_img, center=self.clicked_roi["Tags"][C_TAG], radius=self.radius, color=self.settings['ROI_SELECT_CLR'], thickness=self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_polygon_center(self):
        self.new_vertices = np.array([np.array(v) for k, v in self.new_tags.items() if k != 'Center_tag']).astype(np.int32)
        self.clicked_roi['Tags'] = self.new_tags
        self.clicked_roi[VERTICES] = self.new_vertices
        self.clicked_roi[CENTER_X], self.clicked_roi[CENTER_Y] = self.clicked_roi['Tags']["Center_tag"]
        self.clicked_roi['center'] = self.clicked_roi['Tags']['Center_tag']
        self.temp_img = cv2.polylines(self.temp_img, [self.clicked_roi[VERTICES]], True, color=self.settings['ROI_SELECT_CLR'], thickness=self.clicked_roi['Thickness'],lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)

    def _select_polygon_vertice(self):
        self.clicked_roi['Tags'][self.clicked_tag] = self.new_poly_tag_loc
        self.clicked_roi[VERTICES][self.clicked_tag_id] = np.array([self.new_poly_tag_loc])
        self.clicked_roi['center'] = self.polygon_centroid
        self.clicked_roi['area'] = self.polygon_area
        self.clicked_roi['Tags']['Center_tag'] = tuple(self.polygon_centroid)
        self.temp_img = cv2.polylines(self.temp_img, [self.clicked_roi[VERTICES].astype(np.int32)], True, color=self.clicked_roi['Color BGR'], thickness=self.clicked_roi['Thickness'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][self.clicked_tag], self.clicked_roi['Tags'][self.n_tag_1_id], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.temp_img = cv2.line(self.temp_img, self.clicked_roi['Tags'][self.clicked_tag], self.clicked_roi['Tags'][self.n_tag_2_id], self.settings['ROI_SELECT_CLR'], self.clicked_roi['Thickness'], lineType=self.settings['LINE_TYPE'])
        self.update_image(img=self.temp_img)


    def left_mouse_up(self, event):
        self.edge_selected = False
        if self.clicked_roi is not None:
            self.roi_dict[self.clicked_roi['Name']] = self.clicked_roi
            draw_img = self.get_gridline_copy() if self.rectangle_grid is not None or self.hex_grid is not None else self.original_img.copy()
            self.temp_img = _plot_roi(roi_dict=self.roi_dict, img=draw_img, show_tags=True)
            self.update_image(img=self.temp_img)

    def left_mouse_down(self, event):
        self.click_loc = (event.x, event.y)
        self.clicked_roi, self.clicked_tag = self.find_closest_tag(roi_dict=self.roi_dict, click_coordinate=self.click_loc)
        if self.clicked_roi is not None and self.clicked_tag is not None:
            self.edge_selected = True



    def left_mouse_drag(self, event):
        if self.edge_selected:
            self.x, self.y = (event.x, event.y)
            draw_img = self.get_gridline_copy() if self.rectangle_grid is not None or self.hex_grid is not None else self.original_img.copy()
            self.temp_img = _plot_roi(roi_dict=self.roi_dict, omitted_roi=self.clicked_roi['Name'], img=draw_img)
            if self.clicked_roi['Shape_type'].lower() == ROI_SETTINGS.RECTANGLE.value:
                self.tl_x, self.tl_y = self.clicked_roi['Tags'][TL_TAG]
                self.tr_x, self.tr_y = self.clicked_roi['Tags'][TR_TAG]
                self.br_x, self.br_y = self.clicked_roi['Tags'][BR_TAG]
                self.bl_x, self.bl_y = self.clicked_roi['Tags'][BL_TAG]
                if self.clicked_tag == BR_TAG:
                    self.w, self.h = self.x - self.tl_x, self.y - self.tl_y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_bottom_right()
                if self.clicked_tag == BR_TAG:
                    self.w, self.h = self.x - self.tl_x, self.y - self.tl_y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_bottom_right()
                elif self.clicked_tag == R_TAG:
                    self.w, self.h = self.x - self.tl_x, self.clicked_roi[HEIGHT]
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_right()
                elif self.clicked_tag == B_TAG:
                    self.w, self.h = self.clicked_roi[WIDTH], self.y - self.tl_y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_bottom()
                elif self.clicked_tag == BL_TAG:
                    self.w, self.h = self.tr_x - self.x, self.y - self.tr_y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_bottom_left()
                elif self.clicked_tag == L_TAG:
                    self.w, self.h = self.tr_x - self.x, self.clicked_roi[HEIGHT]
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_left_tag()
                elif self.clicked_tag == T_TAG:
                    self.w, self.h = self.clicked_roi[WIDTH], self.br_y - self.y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_top_tag()
                elif self.clicked_tag == TL_TAG:
                    self.w, self.h = self.br_x - self.x, self.br_y - self.y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_top_left_tag()
                elif self.clicked_tag == TR_TAG:
                    self.w, self.h = self.x - self.bl_x, self.bl_y - self.y
                    if (self.w > 1 < self.h) and (0 <= self.x <= self.img_w) and (0 <= self.y <= self.img_h):
                        self._select_rectangle_top_right_tag()
                elif self.clicked_tag == C_TAG:
                    self.w, self.h = self.clicked_roi[WIDTH], self.clicked_roi[HEIGHT]
                    self.tl_x, self.tl_y = self.x - int(self.w/2), self.y - int(self.h/2)
                    self.tr_x, self.tr_y = self.x + int(self.w/2), self.y - int(self.h/2)
                    self.br_x, self.br_y = self.x + int(self.w/2), self.y + int(self.h/2)
                    if (0 <= self.tl_x) and (0 <= self.tl_y) and (self.tr_x <= self.img_w) and (self.br_y <= self.img_h):
                        self._select_rectangle_center_tag()
            elif self.clicked_roi['Shape_type'].lower() == ROI_SETTINGS.CIRCLE.value:
                if self.clicked_tag == BORDER_TAG:
                    c_x, c_y = self.clicked_roi["Tags"][C_TAG]
                    self.radius = self.clicked_roi["Tags"][C_TAG][0] - self.x
                    self.l_edge, t_edge =  (c_x - self.radius, c_y), (c_x, c_y - self.radius)
                    r_edge, b_edge = (c_x +self. radius, c_y), (c_x, c_y + self.radius)
                    if (0 <= self.l_edge[0]) and (0 <= t_edge[1]) and (r_edge[0] <= self.img_w) and (b_edge[1] <= self.img_h) and (self.radius >= 1):
                        self._select_circle_border()
                if self.clicked_tag == C_TAG:
                    self.radius = self.clicked_roi[RADIUS]
                    self.l_edge, t_edge = (self.x - self.radius, self.y), (self.x, self.y - self.radius)
                    r_edge, b_edge = (self.x + self.radius, self.y), (self.x, self.y + self.radius)
                    if (0 <= self.l_edge[0]) and (0 <= t_edge[1]) and (r_edge[0] <= self.img_w) and (b_edge[1] <= self.img_h):
                        self._select_circle_center()

            elif self.clicked_roi['Shape_type'].lower() == ROI_SETTINGS.POLYGON.value:
                if self.clicked_tag == 'Center_tag':
                    x_diff, y_diff = self.x - self.clicked_roi[CENTER_X], self.y - self.clicked_roi[CENTER_Y]
                    self.new_tags, outside_img_flag = {}, False
                    for tag_name, tag in self.clicked_roi['Tags'].items():
                        t  = (tag[0] + x_diff, tag[1] + y_diff)
                        self.new_tags[tag_name] = t
                        if (0 > t[0]) or (t[0] > self.img_w) or (0 > t[1]) or (t[1] > self.img_h):
                            outside_img_flag = True
                    if not outside_img_flag:
                        self._select_polygon_center()
                else:
                    self.clicked_tag_id = int(self.clicked_tag[-1])
                    click_tag = self.clicked_roi['Tags'][self.clicked_tag]
                    max_tag_id = len(self.clicked_roi["Tags"].keys()) - 2
                    self.n_tag_1_id = f'Tag_{0 if self.clicked_tag_id == max_tag_id else max_tag_id if self.clicked_tag_id == 0 else self.clicked_tag_id - 1}'
                    self.n_tag_2_id = f'Tag_{self.clicked_tag_id - 1 if self.clicked_tag_id == max_tag_id else self.clicked_tag_id + 1}'
                    x_diff, y_diff = self.x - click_tag[0], self.y - click_tag[1]
                    self.new_poly_tag_loc = (click_tag[0] + x_diff, click_tag[1] + y_diff)
                    if (0 < self.new_poly_tag_loc[0]) and (self.new_poly_tag_loc[0] < self.img_w) and (0 < self.new_poly_tag_loc[1]) and (self.new_poly_tag_loc[1] < self.img_h):
                        self.new_vertices = np.copy(self.clicked_roi[VERTICES]).astype(np.int32)
                        self.new_vertices[self.clicked_tag_id] = np.array([self.new_poly_tag_loc])
                        polygon = Polygon(self.new_vertices)
                        self.polygon_area = polygon.area
                        valid_polygon = check_valid_polygon(polygon=polygon, raise_error=False)
                        #self.polygon_centroid = np.array(polygon.centroid).astype(int)
                        self.polygon_centroid = np.array([polygon.centroid.x, polygon.centroid.y]).astype(np.int32)
                        if valid_polygon:
                            self._select_polygon_vertice()
            else:
                pass


#
#
#
#
#
#
#
#
# rectangle_example =  {'Video':             'My_video',
#                       'Shape_type':        ROI_SETTINGS.RECTANGLE.value,
#                       'Name':              'My_rectangle_shape',
#                       'Color name':        'red',
#                       'Color BGR':         (0, 0, 255),
#                       'Thickness':         7,
#                       'Center_X':          175,
#                       'Center_Y':          175,
#                       'topLeftX':          100,
#                       'topLeftY':          100,
#                       'Bottom_right_X':    250,
#                       'Bottom_right_Y':    250,
#                       'width':             150,
#                       'height':            150,
#                       "Tags":             {"Center tag": (175, 175),
#                                            "Top left tag": (100, 100),
#                                            "Bottom right tag": (250, 250),
#                                            "Top right tag": (250, 100),
#                                            "Bottom left tag": (100, 250),
#                                            "Top tag": (175, 100),
#                                            "Right tag": (250, 175),
#                                            "Left tag": (100, 175),
#                                            "Bottom tag": (175, 250)},
#                       'Ear_tag_size':      15}
#
#
# circle_example =      {'Video':            'My_video',
#                        'Shape_type':        ROI_SETTINGS.CIRCLE.value,
#                        'Name':              'My_circle_shape',
#                        'Color name':        'blue',
#                        'Color BGR':          (255, 0, 255),
#                        'Thickness':          7,
#                        'centerX':           400,
#                        'centerY':           400,
#                        'radius':            50,
#                        "Tags":             {"Center tag": (400, 400),
#                                             "Border tag": (350, 400)},
#                        'Ear_tag_size':      15}
#
#
# polygon  = Polygon(np.array([[300, 450], [100, 500], [400, 300], [600, 300], [650, 400]]))
# polygon_vertices  = np.array(polygon.convex_hull.exterior.coords)[1:]
# polygon_center = np.array(polygon.centroid).astype(np.int32)
# polygon_tags = {}
# for cnt, poly in enumerate(polygon_vertices):
#     polygon_tags[f'Tag_{cnt}'] = tuple(poly.astype(np.int32))
# polygon_tags['Center_tag'] = tuple(polygon_center)
#
# polygon_example =      {'Video':                    'My_video',
#                        'Shape_type':                ROI_SETTINGS.POLYGON.value,
#                        'Name':                      'My_polygon_shape',
#                        'Color name':                'yellow',
#                        'Color BGR':                 (0, 255, 255),
#                        'Thickness':                 7,
#                        'Center_X':                  polygon_center[0],
#                        'Center_Y':                  polygon_center[1],
#                        'vertices':                  polygon_vertices,
#                        'center':                    (polygon_center[0], polygon_center[1]),
#                        'area':                      None,
#                        'max_vertice_distance':      None,
#                        "area_cm":                   None,
#                        'Tags':                      polygon_tags,
#                        'Ear_tag_size':              15}
#
# roi_dict = {'My_rectangle_shape': rectangle_example, 'My_circle_shape': circle_example, 'My_polygon_shape': polygon_example}
#
# original_img = read_frm_of_video(video_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0514.mp4")
# frm = read_frm_of_video(video_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0514.mp4")
# for shape_name, shape in roi_dict.items():
#     if shape['Shape_type'] == ROI_SETTINGS.RECTANGLE.value:
#         cv2.rectangle(frm, shape['Tags']['Top left tag'], shape['Tags']['Bottom right tag'], shape['Color BGR'], shape['Thickness'])
#         for tag_name, tag in shape['Tags'].items():
#             cv2.circle(frm, tag, shape['Ear_tag_size'], shape['Color BGR'], -1)
#     if shape['Shape_type'] == ROI_SETTINGS.CIRCLE.value:
#         cv2.circle(frm, center=shape['Tags']['Center tag'], radius=shape['radius'], color=shape['Color BGR'], thickness=shape['Thickness'])
#         for tag_name, tag in shape['Tags'].items():
#             cv2.circle(frm, tag, shape['Ear_tag_size'], shape['Color BGR'], -1)
#     if shape['Shape_type'] == ROI_SETTINGS.POLYGON.value:
#         cv2.polylines(frm, [shape["vertices"].astype(int)], True, color=shape['Color BGR'], thickness=shape['Thickness'])
#         for tag_name, tag in shape['Tags'].items():
#             cv2.circle(frm, tag, shape['Ear_tag_size'], shape['Color BGR'], -1)
#
#
# root = Toplevel()
# root.title(DRAW_FRAME_NAME)
# img_lbl = Label(root, name='img_lbl')
# img_lbl.pack()
#
# img_rgb = cv2.cvtColor(frm, cv2.COLOR_BGR2RGB)
# pil_image = Image.fromarray(img_rgb)
# tk_image = ImageTk.PhotoImage(pil_image)
# img_lbl.configure(image=tk_image)
# img_lbl.image = tk_image
#
# _ = InteractiveROIModifier(img_window=root, roi_dict=roi_dict, original_img=original_img)
# root.mainloop()


import os
from tkinter import *
from typing import List, Optional, Union

from simba.data_processors.cue_light_clf_statistics import CueLightClfAnalyzer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.checks import check_if_dir_exists, check_int, check_valid_lst
from simba.utils.enums import Links
from simba.utils.errors import NoDataError, NoFilesFoundError, NoROIDataError
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)


class CueLightClfAnalyzerPopUp(ConfigReader, PopUpMixin):

    """
    :example:
    >>> CueLightClfAnalyzerPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini", cue_light_names=['cl'])
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 cue_light_names: List[str],
                 data_dir: Optional[Union[str, os.PathLike]] = None):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True)
        check_valid_lst(data=cue_light_names, source=self.__class__.__name__, valid_dtypes=(str,), min_len=1, raise_error=True)
        self.read_roi_data()
        self.cue_light_names = cue_light_names
        missing_cue_lights = [x for x in cue_light_names if x not in self.roi_names]
        if len(missing_cue_lights) > 0: raise NoROIDataError(msg=f'{len(missing_cue_lights)} cue-lights are not drawn in the SimBA project: cannot be found in the {self.roi_coordinates_path} file', source=self.__class__.__name__)

        if data_dir is None:
            self.data_dir = self.cue_lights_data_dir
        else:
            check_if_dir_exists(in_dir=data_dir)
            self.data_dir = data_dir

        self.data_paths = find_files_of_filetypes_in_directory(directory=self.data_dir, extensions=[f'.{self.file_type}'], raise_error=False, as_dict=True)
        file_cnt = len(list(self.data_paths.keys())) if type(self.data_paths) == dict else len(self.data_paths)
        if file_cnt == 0:
            raise NoFilesFoundError(msg=f'No data files exist in the {self.data_dir} directory: ANALYZE CUE LIGHT DATA before performing cue light classification analysis.', source=self.__class__.__name__)

        PopUpMixin.__init__(self, size=(750, 300), title="SIMBA CUE LIGHT CLASSIFICATION ANALYZER", icon='clf_2')
        self.clf_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="CLASSIFIERS", icon_name='clf_2', icon_link=Links.CUE_LIGHTS.value)
        self.clf_vars = {}
        for cnt, clf in enumerate(self.clf_names):
            clf_cb, clf_var = SimbaCheckbox(parent=self.clf_frm, txt=clf, val=True)
            clf_cb.grid(row=cnt, column=0, sticky=NW)
            self.clf_vars[clf] = clf_var
        self.clf_frm.grid(row=0, column=0, sticky=NW)

        self.settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings', icon_link=Links.CUE_LIGHTS.value)
        self.pre_window_eb = Entry_Box(parent=self.settings_frm, fileDescription='PRE CUE LIGHT WINDOW (S): ', entry_box_width=30, value=0, validation='numeric', labelwidth="40", justify='center')
        self.post_window_eb = Entry_Box(parent=self.settings_frm, fileDescription='POST CUE LIGHT WINDOW (S): ', entry_box_width=30, value=0, validation='numeric', labelwidth="40", justify='center')

        self.settings_frm.grid(row=1, column=0, sticky=NW)
        self.pre_window_eb.grid(row=0, column=0, sticky=NW)
        self.post_window_eb.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()


    def run(self):
        pre_window = self.pre_window_eb.entry_get.strip()
        post_window = self.post_window_eb.entry_get.strip()
        check_int(name='PRE CUE LIGHT WINDOW (S): ', value=pre_window, min_value=0, raise_error=False)
        check_int(name='POST CUE LIGHT WINDOW (S): ', value=post_window, min_value=0, raise_error=False)
        clf_names = [x for x in self.clf_vars.keys() if self.clf_vars[x].get()]
        if len(clf_names) == 0:
            raise NoDataError(msg='Zero classifiers checked. Check at least one classifier.', source=self.__class__.__name__)
        analyzer = CueLightClfAnalyzer(config_path=self.config_path,
                                       cue_light_names=self.cue_light_names,
                                       clf_names=clf_names,
                                       pre_window=int(pre_window),
                                       post_window=int(post_window))
        analyzer.run()
        analyzer.save()


#CueLightClfAnalyzerPopUp(config_path=r"C:\troubleshooting\cue_light\t1\project_folder\project_config.ini", cue_light_names=['MY_CUE_LIGHT'])



__author__ = "Simon Nilsson"

import glob
import os
from typing import Any, Dict, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.unsupervised_mixin import UMLMixin
from simba.unsupervised.enums import Clustering, UMLOptions, Unsupervised
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty,
                                check_if_keys_exist_in_dict, check_instance,
                                check_int, check_str, check_that_column_exist,
                                check_valid_lst)
from simba.utils.enums import Formats, Options
from simba.utils.printing import stdout_success
from simba.utils.read_write import read_pickle


class GridSearchVisualizer(UMLMixin):
    """
    Visualize grid-searched latent spaces in .png format.

    .. image:: _static/img/GridSearchVisualizer.png
       :width: 800
       :align: center

    :param model_dir: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param save_dir: directory holding one or more unsupervised results in pickle ``data_map.yaml`` format.
    :param settings: User-defined image attributes (e.g., continous and catehorical palettes)

    :example:
    >>> settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'CONTINUOUS_PALETTE': 'magma', 'SCATTER_SIZE': 10}
    >>> visualizer = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models_042023', save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images', settings=settings)
    >>> visualizer.continuous_visualizer(continuous_vars=['START_FRAME'])
    >>> visualizer.categorical_visualizer(categoricals=['CLUSTER'])
    """

    def __init__(
        self,
        model_dir: Union[str, os.PathLike],
        save_dir: Union[str, os.PathLike],
        settings: Dict[str, Any],
    ):
        super().__init__()
        check_if_dir_exists(in_dir=save_dir)
        check_if_dir_exists(in_dir=model_dir)
        self.save_dir, self.settings, self.model_dir = save_dir, settings, model_dir
        self.data_paths = glob.glob(model_dir + f"/*.{Formats.PICKLE.value}")
        check_if_keys_exist_in_dict(
            data=self.settings,
            key=["CATEGORICAL_PALETTE", "CONTINUOUS_PALETTE", "SCATTER_SIZE"],
            name=f"{self.__class__.__name__} settings",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.data_paths,
            error_msg=f"SIMBA ERROR: No pickle files found in {model_dir}",
        )
        check_int(
            name=f"{self.__class__.__name__} scatter size",
            value=self.settings["SCATTER_SIZE"],
            min_value=1,
        )
        check_str(
            name=f"{self.__class__.__name__} categorical palette",
            value=self.settings["CATEGORICAL_PALETTE"],
            options=Options.PALETTE_OPTIONS_CATEGORICAL.value,
        )
        check_str(
            name=f"{self.__class__.__name__} continuous palette",
            value=self.settings["CONTINUOUS_PALETTE"],
            options=Options.PALETTE_OPTIONS.value,
        )

    def __extract_plot_data(self, data: dict):
        embedding_data = pd.DataFrame(
            data[Unsupervised.DR_MODEL.value][Unsupervised.MODEL.value].embedding_,
            columns=["X", "Y"],
        )
        bouts_data = data[Unsupervised.DATA.value][
            Unsupervised.BOUTS_FEATURES.value
        ].reset_index()
        target_data = data[Unsupervised.DATA.value][Unsupervised.BOUTS_TARGETS.value]
        cluster_data = pd.DataFrame(
            data[Clustering.CLUSTER_MODEL.value][Unsupervised.MODEL.value]
            .labels_.reshape(-1, 1)
            .astype(np.int8),
            columns=["CLUSTER"],
        )
        data = pd.concat(
            [embedding_data, bouts_data, target_data, cluster_data], axis=1
        )
        return data.loc[:, ~data.columns.duplicated()].copy()

    def categorical_visualizer(self, categorical_vars: List[str]):
        check_valid_lst(
            data=categorical_vars,
            source=self.__class__.__name__,
            valid_dtypes=(str,),
            min_len=1,
        )
        print(
            f"Creating {len(categorical_vars)} categorical plot(s) from {len(self.data_paths)} data files..."
        )
        for file_cnt, file_path in enumerate(self.data_paths):
            v = read_pickle(data_path=file_path)
            check_if_keys_exist_in_dict(
                data=v,
                key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                name=file_path,
            )
            data = self.__extract_plot_data(data=v)
            for variable in categorical_vars:
                check_that_column_exist(
                    df=data,
                    column_name=variable,
                    file_name=v[Unsupervised.DR_MODEL.value][
                        Unsupervised.HASHED_NAME.value
                    ],
                )
                save_path = os.path.join(
                    self.save_dir,
                    f"{v[Clustering.CLUSTER_MODEL.value][Unsupervised.HASHED_NAME.value]}_{variable}.png",
                )
                _ = PlottingMixin.categorical_scatter(
                    data=data,
                    columns=("X", "Y", variable),
                    palette=self.settings["CATEGORICAL_PALETTE"],
                    size=self.settings["SCATTER_SIZE"],
                    save_path=save_path,
                    show_box=False,
                )
                stdout_success(msg=f"Saved {save_path}...")
                plt.close("all")
        self.timer.stop_timer()
        stdout_success(
            msg=f"{int(len(categorical_vars) * len(self.data_paths))} categorical plot(s) created.",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def continuous_visualizer(self, continuous_vars: List[str]):
        check_valid_lst(
            data=continuous_vars,
            source=self.__class__.__name__,
            valid_dtypes=(str,),
            min_len=1,
        )
        print(
            f"Creating {len(continuous_vars)} categorical plot(s) from {len(self.data_paths)} data files..."
        )
        for file_cnt, file_path in enumerate(self.data_paths):
            v = read_pickle(data_path=file_path)
            check_if_keys_exist_in_dict(
                data=v,
                key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                name=file_path,
            )
            data = self.__extract_plot_data(data=v)
            for variable in continuous_vars:
                check_that_column_exist(
                    df=data,
                    column_name=variable,
                    file_name=v[Unsupervised.DR_MODEL.value][
                        Unsupervised.HASHED_NAME.value
                    ],
                )
                save_path = os.path.join(
                    self.save_dir,
                    f"{v[Clustering.CLUSTER_MODEL.value][Unsupervised.HASHED_NAME.value]}_{variable}.png",
                )
                _ = PlottingMixin.continuous_scatter(
                    data=data,
                    columns=("X", "Y", variable),
                    palette=self.settings["CONTINUOUS_PALETTE"],
                    size=self.settings["SCATTER_SIZE"],
                    save_path=save_path,
                    show_box=False,
                )
                plt.close("all")
                stdout_success(msg=f"Saved {save_path}...")
        self.timer.stop_timer()
        stdout_success(
            msg=f"{int(len(continuous_vars) * len(self.data_paths))} continuous plot(s) created.",
            elapsed_time=self.timer.elapsed_time_str,
        )


# #
# settings = {'CATEGORICAL_PALETTE': 'tab20', 'CONTINUOUS_PALETTE': 'magma', 'SCATTER_SIZE': 10}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/clusters',
#                             save_dir='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/cluster_vis',
#                             settings=settings)
# test.categorical_visualizer(categorical_vars=['CLUSTER'])
# test.continuous_visualizer(continuous_vars=['PROBABILITY'])


# settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'CONTINUOUS_PALETTE': 'magma', 'SCATTER_SIZE': 10}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/clustering_test_2',
#                             save_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/cluster_viz_2',
#                             settings=settings)
# test.categorical_visualizer(categorical_vars=['CLASSIFIER'])


# settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'CONTINUOUS_PALETTE': 'magma', 'SCATTER_SIZE': 10}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/cluster_mdls',
#                             save_dir='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/cluster_vis',
#                             settings=settings)
# #test.categorical_visualizer(categorical_vars=['VIDEO', 'CLUSTER'])
# test.continuous_visualizer(continuous_vars=['START_FRAME'])

# settings = {'PALETTE': 'Pastel1'}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models',
#                             save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images',
#                             settings=settings)
# test.cluster_visualizer()


# settings = {'CATEGORICAL_PALETTE': 'Pastel1', 'SCATTER_SIZE': 10}
# test = GridSearchVisualizer(model_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/cluster_models_042023',
#                             save_dir='/Users/simon/Desktop/envs/troubleshooting/unsupervised/images',
#                             settings=settings)
# #test.continuous_visualizer(continuous_vars=['START_FRAME'])
# test.categorical_visualizer(categoricals=['CLUSTER'])



__author__ = "Simon Nilsson"

import os
import shutil
from typing import Any, Dict, List, Optional, Union

import pandas as pd

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_keys_exist_in_dict, check_instance,
                                check_int, check_str)
from simba.utils.errors import FileExistError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext,
                                    get_number_of_header_columns_in_df)

DLC_ = 'DLC_'
DeepCut = 'DeepCut'

def import_dlc_csv(config_path: Union[str, os.PathLike], source: str) -> List[str]:
    """
    Import file or folder of  DLC pose-estimation CSV files to SimBA project.
    Returns list of file paths that has been imported.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter str source: path to file or folder containing DLC pose-estimation CSV files
    :return List[str]: Paths to location of imported files.

    :example:
    >>> import_dlc_csv(config_path='project_folder/project_config.ini', source='CSV_import/Together_1.csv')
    >>> ['project_folder/csv/input_csv/Together_1.csv']
    """

    check_file_exist_and_readable(file_path=config_path)
    conf = ConfigReader(config_path=config_path, read_video_info=False)
    original_file_name_dir = os.path.join(conf.input_csv_dir, "original_filename")
    if not os.path.exists(original_file_name_dir): os.makedirs(original_file_name_dir)
    prev_imported_file_paths = find_files_of_filetypes_in_directory(directory=conf.input_csv_dir, extensions=[f'.{conf.file_type}'], raise_warning=False, raise_error=False)
    prev_imported_file_names = [get_fn_ext(x)[1] for x in prev_imported_file_paths]
    if os.path.isdir(source):
        new_data_paths = find_files_of_filetypes_in_directory(directory=source, extensions=['.csv'], raise_warning=False, raise_error=True)
    elif os.path.isfile(source):
        new_data_paths = [source]
    else:
        raise InvalidInputError(msg=f'{source} is not a valid data directory path or file path.', source=import_dlc_csv.__name__)

    imported_file_paths = []
    for file_cnt, file_path in enumerate(new_data_paths):
        video_timer = SimbaTimer(start=True)
        check_file_exist_and_readable(file_path=file_path)
        _, video_name, file_ext = get_fn_ext(filepath=file_path)
        if DLC_ in video_name:
            new_file_name = video_name.split(DLC_)[0] + ".csv"
        elif DeepCut in video_name:
            new_file_name = video_name.split(DeepCut)[0] + ".csv"
        else:
            new_file_name = video_name.split(".")[0] + ".csv"
        new_file_name_wo_ext = new_file_name.split(".")[0]
        video_basename = os.path.basename(file_path)
        print(f"Importing {video_name} to SimBA project...")
        if new_file_name_wo_ext in prev_imported_file_names:
            raise FileExistError(f"SIMBA IMPORT ERROR: {new_file_name} already exist in project in the directory {conf.input_csv_dir}. Remove file from project or rename imported video file name before importing.")
        shutil.copy(file_path, conf.input_csv_dir)
        shutil.copy(file_path, original_file_name_dir)
        os.rename(os.path.join(conf.input_csv_dir, video_basename), os.path.join(conf.input_csv_dir, new_file_name))
        df = pd.read_csv(os.path.join(conf.input_csv_dir, new_file_name))
        header_cols = get_number_of_header_columns_in_df(df=df)
        if header_cols == 3:
            df = df.iloc[1:]
        if conf.file_type == "parquet":
            df = pd.read_csv(os.path.join(conf.input_csv_dir, video_basename))
            df = df.apply(pd.to_numeric, errors="coerce")
            df.to_parquet(os.path.join(conf.input_csv_dir, new_file_name))
            os.remove(os.path.join(conf.input_csv_dir, video_basename))
        if conf.file_type == "csv":
            df.to_csv(os.path.join(conf.input_csv_dir, new_file_name), index=False)
        imported_file_paths.append(os.path.join(conf.input_csv_dir, new_file_name))
        video_timer.stop_timer()
        print(f"Pose-estimation data for video {video_name} imported to SimBA project (elapsed time: {video_timer.elapsed_time_str}s)...")
    return imported_file_paths

def import_dlc_csv_data(config_path: Union[str, os.PathLike],
                        data_path: Union[str, os.PathLike],
                        interpolation_settings: Optional[Dict[str, Any]] = None,
                        smoothing_settings: Optional[Dict[str, Any]] = None) -> None:

    """
    Import multiple DLC CSV tracking files to SimBA project and apply specified interpolation and smoothing
    parameters to the imported data.

    :param Union[str, os.PathLike] config_path: Path to SimBA config file in ConfigParser format.
    :param Union[str, os.PathLike] data_path: Path to directory holding DLC pose-estimation data in CSV format, or path to a single CSV file with DLC pose-estimation data.
    :param Optional[Dict[str, Any]] interpolation_settings: Dictionary holding settings for interpolation.
    :param Optional[Dict[str, Any]] smoothing_settings: Dictionary holding settings for smoothing.
    :return None:

    :example:
    >>> interpolation_settings = {'type': 'body-parts', 'method': 'linear'}
    >>> smoothing_settings = None #{'time_window': 500, 'method': 'savitzky-golay'}
    >>> import_dlc_csv_data(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', data_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/new_data', interpolation_settings=interpolation_settings, smoothing_settings=smoothing_settings)

    """

    timer = SimbaTimer(start=True)
    check_file_exist_and_readable(file_path=config_path)
    if (not os.path.isdir(data_path)) and (not os.path.isfile(data_path)):
        raise InvalidInputError(msg=f'{data_path} is not a valid data directory path or file path.', source=import_dlc_csv.__name__)
    if interpolation_settings != None:
        check_instance(source=f'{import_dlc_csv_data.__name__} interpolation_settings', accepted_types=(dict,), instance=interpolation_settings)
        check_if_keys_exist_in_dict(data=interpolation_settings, key=['type', 'method'])
        check_str(name='type', value=interpolation_settings['type'].lower(), options=['animals', 'body-parts'])
        check_str(name='method', value=interpolation_settings['method'].lower(), options=['nearest', 'linear', 'quadratic'])
    if smoothing_settings != None:
        check_instance(source=f'{import_dlc_csv_data.__name__} smoothing_settings', accepted_types=(dict,), instance=smoothing_settings)
        check_if_keys_exist_in_dict(data=smoothing_settings, key=['time_window', 'method'])
        check_int(name='time_window', value=smoothing_settings['time_window'], min_value=1)
        check_str(name='method', value=smoothing_settings['method'].lower(), options=['savitzky-golay', 'gaussian'])

    imported_file_paths = import_dlc_csv(config_path=config_path, source=data_path)
    if interpolation_settings != None:
        interpolator = Interpolate(config_path=config_path, data_path=imported_file_paths, type=interpolation_settings['type'], method=interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
        interpolator.run()
    if smoothing_settings != None:
        smoother = Smoothing(config_path=config_path, data_path=imported_file_paths, time_window=smoothing_settings['time_window'], method=smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
        smoother.run()
    timer.stop_timer()
    stdout_success(msg=f"Imported {len(imported_file_paths)} pose estimation file(s) to directory project cvs/input_csv", elapsed_time=timer.elapsed_time_str)



__author__ = "Simon Nilsson"

import json
import os
from typing import List, Optional, Union

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.unsupervised.enums import Clustering, UMLOptions, Unsupervised
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict,
                                check_valid_extension)
from simba.utils.enums import Formats
from simba.utils.errors import InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_pickle, write_pickle)

CLUSTERER_PARAMETERS = "CLUSTERER HYPER-PARAMETERS"
DIMENSIONALITY_REDUCTION_PARAMETERS = "DIMENSIONALITY REDUCTION HYPER-PARAMETERS"
SCALER = "SCALER"
SCALED_DATA = "SCALED DATA"
LOW_VARIANCE_FIELDS = "LOW VARIANCE FIELDS"
FEATURE_NAMES = "FEATURE NAMES"
FRAME_FEATURES = "FRAME FEATURES"
FRAME_POSE = "FRAME POSE"
FRAME_TARGETS = "FRAME TARGETS"
BOUTS_FEATURES = "BOUTS FEATURES"
BOUTS_TARGETS = "BOUTS TARGETS"
BOUTS_DIM_CORDS = "BOUTS DIMENSIONALITY REDUCTION DATA"
BOUTS_CLUSTER_LABELS = "BOUTS CLUSTER LABELS"


class DataExtractor(UnsupervisedMixin, ConfigReader):
    """
    Extracts human-readable data from directory of pickles or single pickle file that holds unsupervised analyses.

    :param config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param data_path: path to pickle holding unsupervised results in ``data_map.yaml`` format.
    :param data_type: The type of data to extract. E.g., CLUSTERER_PARAMETERS, DIMENSIONALITY_REDUCTION_PARAMETERS, SCALER, SCALED_DATA, LOW_VARIANCE_FIELDS, FEATURE_NAMES, FRAME_FEATURES, FRAME_POSE, FRAME_TARGET, BOUTS_FEATURES, BOUTS_TARGETS, BOUTS_DIM_CORDS
    :param settings: User-defined parameters for data extraction.

    :example:
    >>> extractor = DataExtractor(data_path='unsupervised/cluster_models/awesome_curran.pickle', data_type=['BOUTS_TARGETS'], settings=None, config_path='unsupervised/project_folder/project_config.ini')
    >>> extractor.run()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_path: Union[str, os.PathLike],
        data_types: List[str],
        settings: Optional[dict] = None,
    ):

        check_file_exist_and_readable(file_path=config_path)
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        if os.path.isdir(data_path):
            check_if_dir_exists(in_dir=data_path)
            self.data_paths = find_files_of_filetypes_in_directory(
                directory=data_path,
                extensions=[f".{Formats.PICKLE.value}"],
                raise_error=True,
            )
        else:
            check_valid_extension(
                path=data_path, accepted_extensions=Formats.PICKLE.value
            )
            self.data_paths = [data_path]
        invalid_dtypes = list(set(data_types) - set(UMLOptions.DATA_TYPES.value))
        if len(data_types) == 0:
            raise InvalidInputError(
                msg=f"data_types is an empty list. Accepted options: {UMLOptions.DATA_TYPES.value}.",
                source=self.__class__.__name__,
            )
        if len(invalid_dtypes) > 0:
            raise InvalidInputError(
                msg=f"Found invalid data types: {invalid_dtypes}. Accepted: {UMLOptions.DATA_TYPES.value}.",
                source=self.__class__.__name__,
            )
        self.save_dir = os.path.join(
            self.logs_path, f"extracted_unsupervised_model_data_{self.datetime}"
        )
        if not os.path.isdir(self.save_dir):
            os.makedirs(self.save_dir)
        self.settings, self.data_types = settings, data_types

    def run(self):
        for file_cnt, file_path in enumerate(self.data_paths):
            print(f"Processing file {file_cnt+1}/{len(self.data_paths)}...")
            file_timer = SimbaTimer(start=True)
            v = read_pickle(data_path=file_path)
            mdl_name = get_fn_ext(filepath=file_path)[1]
            save_subdir = os.path.join(self.save_dir, get_fn_ext(filepath=file_path)[1])
            if not os.path.isdir(save_subdir):
                os.makedirs(save_subdir)
            for data_type in self.data_types:
                if data_type == CLUSTERER_PARAMETERS:
                    check_if_keys_exist_in_dict(
                        data=v, key=[Clustering.CLUSTER_MODEL.value], name=file_path
                    )
                    save_path = os.path.join(
                        save_subdir, f"cluster_parameters_{mdl_name}.json"
                    )
                    json.dump(
                        v[Clustering.CLUSTER_MODEL.value][
                            Unsupervised.PARAMETERS.value
                        ],
                        open(save_path, "w"),
                        indent=4,
                        sort_keys=True,
                    )
                    print(
                        f"Saved cluster parameters for model {mdl_name} at {save_path}..."
                    )

                elif data_type == DIMENSIONALITY_REDUCTION_PARAMETERS:
                    check_if_keys_exist_in_dict(
                        data=v, key=[Unsupervised.DR_MODEL.value], name=file_path
                    )
                    save_path = os.path.join(
                        save_subdir,
                        f"dimensionality_reduction_parameters_{mdl_name}.json",
                    )
                    json.dump(
                        v[Unsupervised.DR_MODEL.value][Unsupervised.PARAMETERS.value],
                        open(save_path, "w"),
                        indent=4,
                        sort_keys=True,
                    )
                    print(
                        f"Saved dimension reduction parameters for model {mdl_name} at {save_path} ..."
                    )
                #
                elif data_type == SCALER:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.METHODS.value],
                        name=file_path,
                    )
                    mdl_name = v[Unsupervised.DR_MODEL.value][
                        Unsupervised.HASHED_NAME.value
                    ]
                    save_path = os.path.join(save_subdir, f"scaler_{mdl_name}.pickle")
                    write_pickle(
                        data=v[Unsupervised.METHODS.value][Unsupervised.SCALER.value],
                        save_path=save_path,
                    )
                    print(f"Saved scaler for model {mdl_name} at {save_path}...")

                elif data_type == SCALED_DATA:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.METHODS.value],
                        name=file_path,
                    )
                    save_path = os.path.join(save_subdir, f"scaled_data_{mdl_name}.csv")
                    v[Unsupervised.METHODS.value][
                        Unsupervised.SCALED_DATA.value
                    ].to_csv(save_path)
                    print(f"Saved scaled data for {mdl_name} model at {save_path}...")

                elif data_type == LOW_VARIANCE_FIELDS:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.METHODS.value],
                        name=file_path,
                    )
                    save_path = os.path.join(
                        save_subdir, f"low_variance_fields_{mdl_name}.csv"
                    )
                    out_df = pd.DataFrame(
                        data=v[Unsupervised.METHODS.value][
                            Unsupervised.LOW_VARIANCE_FIELDS.value
                        ],
                        columns=["FIELD_NAMES"],
                    )
                    out_df.to_csv(save_path)
                    print(
                        f"Saved low variance fields for model {mdl_name} at {save_path}..."
                    )

                elif data_type == FEATURE_NAMES:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.METHODS.value],
                        name=file_path,
                    )
                    save_path = os.path.join(
                        save_subdir, f"feature_names_{mdl_name}.csv"
                    )
                    out_df = pd.DataFrame(
                        data=v[Unsupervised.METHODS.value][
                            Unsupervised.FEATURE_NAMES.value
                        ],
                        columns=["FIELD_NAMES"],
                    )
                    out_df.to_csv(save_path)
                    print(f"Feature names saved for model {mdl_name} at {save_path}...")

                elif data_type == FRAME_FEATURES:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"frame_wise_features_{mdl_name}.csv"
                    )
                    v[Unsupervised.DATA.value][
                        Unsupervised.FRAME_FEATURES.value
                    ].to_csv(save_path)
                    print(
                        f"Saved frame-wise features for model {mdl_name} at {save_path}..."
                    )

                elif data_type == FRAME_POSE:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"frame_wise_pose_estimation_{mdl_name}.csv"
                    )
                    v[Unsupervised.DATA.value][Unsupervised.FRAME_POSE.value].to_csv(
                        save_path
                    )
                    print(
                        f"Frame-wise pose saved for model {mdl_name} at {save_path}..."
                    )

                elif data_type == FRAME_TARGETS:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"frame_wise_target_data_{mdl_name}.csv"
                    )
                    v[Unsupervised.DATA.value][Unsupervised.FRAME_TARGETS.value].to_csv(
                        save_path
                    )
                    print(
                        f"Saved frame-wise target data for model {mdl_name} at {save_path}..."
                    )

                elif data_type == BOUTS_FEATURES:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"bout_features_data_{mdl_name}.csv"
                    )
                    v[Unsupervised.DATA.value][
                        Unsupervised.BOUTS_FEATURES.value
                    ].to_csv(save_path)
                    print(
                        f"Saved bout features data  for model {mdl_name} at {save_path}..."
                    )

                elif data_type == BOUTS_TARGETS:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value, Unsupervised.DATA.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"bout_targets_{mdl_name}.csv"
                    )
                    v[Unsupervised.DATA.value][Unsupervised.BOUTS_TARGETS.value].to_csv(
                        save_path
                    )
                    print(
                        f"Saved bout target data for model {mdl_name} at {save_path}..."
                    )

                elif data_type == BOUTS_DIM_CORDS:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[Unsupervised.DR_MODEL.value],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"bout_dim_cords_{mdl_name}.csv"
                    )
                    mdl_data = v[Unsupervised.DR_MODEL.value][
                        Unsupervised.MODEL.value
                    ].embedding_.astype(np.float32)
                    idx = v[Unsupervised.METHODS.value][
                        Unsupervised.SCALED_DATA.value
                    ].index
                    mdl_data = pd.DataFrame(
                        data=mdl_data, index=idx, columns=["X", "Y"]
                    )
                    mdl_data.to_csv(save_path)
                    print(
                        f"Saved bout dimensionality reduction data for model {mdl_name} at {save_path}..."
                    )

                elif data_type == BOUTS_CLUSTER_LABELS:
                    check_if_keys_exist_in_dict(
                        data=v,
                        key=[
                            Unsupervised.DR_MODEL.value,
                            Clustering.CLUSTER_MODEL.value,
                        ],
                        name=file_path,
                    )

                    save_path = os.path.join(
                        save_subdir, f"bout_cluster_labels_{mdl_name}.csv"
                    )
                    mdl_data = (
                        v[Clustering.CLUSTER_MODEL.value][Unsupervised.MODEL.value]
                        .labels_.astype(np.int64)
                        .reshape(-1, 1)
                    )
                    idx = v[Unsupervised.METHODS.value][
                        Unsupervised.SCALED_DATA.value
                    ].index
                    mdl_data = pd.DataFrame(data=mdl_data, index=idx, columns=["LABEL"])
                    mdl_data.to_csv(save_path)
                    print(
                        f"Saved bout cluster labels for model {mdl_name} at {save_path}..."
                    )

                else:
                    raise InvalidInputError(
                        msg=f"Invalid datatype {data_type}.",
                        source=self.__class__.__name__,
                    )

            file_timer.stop_timer()
            stdout_success(
                msg=f"{mdl_name} model data extraction complete",
                elapsed_time=file_timer.elapsed_time,
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Data for {len(self.data_paths)} model(s) extracted",
            elapsed_time=self.timer.elapsed_time,
        )


# test = DataExtractor(data_path='/Users/simon/Desktop/envs/simba/troubleshooting/NG_Unsupervised/project_folder/cluster_mdls/hopeful_khorana.pickle',
#                      data_types=['CLUSTERER HYPER-PARAMETERS'],
#                      settings=None,
#                      config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini')
#
# test.run()
